#!/usr/bin/env python
# encoding: utf-8
"""
ModelClasses.py

Created by José Sánchez-Gallego on 23 Jul 2015.
Licensed under a 3-clause BSD license.

Revision history:
    23 Jul 2015 J. Sánchez-Gallego
      Initial version
    21 Feb 2016 J. Sánchez-Gallego
      Rewritten as classes derived from declarative base.

"""

from __future__ import division, print_function

import itertools
import math
import re
import shutil

import numpy as np
from marvin.core.caching_query import RelationshipCache
from marvin.db.database import db
from sqlalchemy import Float, ForeignKeyConstraint, case, cast, func
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.inspection import inspect as sa_inspect
from sqlalchemy.orm import backref, configure_mappers, relationship


try:
    import cStringIO as StringIO
except ImportError:
    from io import StringIO


Base = db.Base


# The softening parameter for asinh magnitudes
bb = {'u': 1.4e-10,
      'g': 0.9e-10,
      'r': 1.2e-10,
      'i': 1.8e-10,
      'z': 7.4e-10}


def cameliseClassname(tableName):
    """Produce a camelised class name."""

    return str(tableName[0].upper() +
               re.sub(r'_([a-z])',
               lambda m: m.group(1).upper(), tableName[1:]))


def ClassFactory(name, tableName, BaseClass=db.Base, fks=None):
    tableArgs = [{'autoload': True, 'schema': 'mangasampledb'}]
    if fks:
        for fk in fks:
            tableArgs.insert(0, ForeignKeyConstraint([fk[0]], [fk[1]]))

    newclass = type(
        name, (BaseClass,),
        {'__tablename__': tableName,
         '__table_args__': tuple(tableArgs)})

    return newclass


class MangaTarget(Base):
    __tablename__ = 'manga_target'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    def __repr__(self):
        return '<MangaTarget (pk={0}, mangaid={1})>'.format(self.pk,
                                                            self.mangaid)


class Anime(Base):
    __tablename__ = 'anime'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    def __repr__(self):
        return '<Anime (pk={0}, anime={1})>'.format(self.pk, self.anime)


class Character(Base):
    __tablename__ = 'character'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    target = relationship(MangaTarget, backref='character', uselist=False)
    anime = relationship(Anime, backref='characters')

    def __repr__(self):
        return '<Character (pk={0}, name={1})>'.format(self.pk, self.name)

    def savePicture(self, path):
        """Saves the picture blob to disk."""

        buf = StringIO(self.picture)
        with open(path, 'w') as fd:
            buf.seek(0)
            shutil.copyfileobj(buf, fd)

        return buf


class Catalogue(Base):
    __tablename__ = 'catalogue'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    @property
    def isCurrent(self):
        return self.currentCatalogue is not None

    def __repr__(self):
        return ('<Catalogue (pk={0}), catalogue={1}, version={2}, current={3}>'
                .format(self.pk, self.catalogue_name, self.version,
                        self.isCurrent))


class CurrentCatalogue(Base):
    __tablename__ = 'current_catalogue'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    catalogue = relationship(
        'Catalogue', backref=backref('currentCatalogue', uselist=False))

    def __repr__(self):
        return '<CurrentCatalogue (pk={0})>'.format(self.pk)


class MangaTargetToMangaTarget(Base):
    __tablename__ = 'manga_target_to_manga_target'
    __table_args__ = {'autoload': True, 'schema': 'mangasampledb'}

    def __repr__(self):
        return '<MangaTargetToMangaTarget (pk={0})>'.format(self.pk)


class NSA(Base):
    __tablename__ = 'nsa'
    __table_args__ = (
        ForeignKeyConstraint(['catalogue_pk'], ['mangasampledb.catalogue.pk']),
        {'autoload': True, 'schema': 'mangasampledb'})

    def __repr__(self):
        return '<NSA (pk={0}, nsaid={1})>'.format(self.pk, self.nsaid)


class MangaTargetToNSA(Base):
    __tablename__ = 'manga_target_to_nsa'
    __table_args__ = (
        ForeignKeyConstraint(['manga_target_pk'],
                             ['mangasampledb.manga_target.pk']),
        ForeignKeyConstraint(['nsa_pk'], ['mangasampledb.nsa.pk']),
        {'autoload': True, 'schema': 'mangasampledb'})

    def __repr__(self):
        return '<MangaTargetToNSA (pk={0})>'.format(self.pk)


# Relationship between NSA and MangaTarget
NSA.mangaTargets = relationship(
    MangaTarget, backref='NSA_objects', secondary=MangaTargetToNSA.__table__)

# Now we create the remaining tables.
insp = sa_inspect(db.engine)
schemaName = 'mangasampledb'
allTables = insp.get_table_names(schema=schemaName)

done_names = list(db.Base.metadata.tables.keys())
for tableName in allTables:
    if schemaName + '.' + tableName in done_names:
        continue
    className = str(tableName).upper()

    newClass = ClassFactory(
        className, tableName,
        fks=[('catalogue_pk', 'mangasampledb.catalogue.pk')])
    newClass.catalogue = relationship(
        Catalogue, backref='{0}_objects'.format(tableName))
    locals()[className] = newClass
    done_names.append(schemaName + '.' + tableName)

    if 'manga_target_to_' + tableName in allTables:
        relationalTableName = 'manga_target_to_' + tableName
        relationalClassName = 'MangaTargetTo' + tableName.upper()
        newRelationalClass = ClassFactory(
            relationalClassName, relationalTableName,
            fks=[('manga_target_pk', 'mangasampledb.manga_target.pk'),
                 ('nsa_pk', 'mangasampledb.nsa.pk')])

        locals()[relationalClassName] = newRelationalClass
        done_names.append(schemaName + '.' + relationalTableName)

        newClass.mangaTargets = relationship(
            MangaTarget, backref='{0}_objects'.format(tableName),
            secondary=newRelationalClass.__table__)


def HybridMag(flux_parameter, band, index=None):
    """Returns a hybrid property describing an asinh magnitude.

    ``flux_parameter`` must be a column with a flux in nanomaggies. ``band`` is
    the band name, to determine the softening parameter. If ``flux_parameter``
    is and array, ``index`` defines the position of ``band`` within the array.

    """

    @hybrid_property
    def hybridMag(self):
        if index is not None:
            flux = getattr(self, flux_parameter)[index]
        else:
            flux = getattr(self, flux_parameter)

        flux *= 1e-9  # From nanomaggies to maggies
        bb_band = bb[band]
        asinh_mag = -2.5 / np.log(10) * (np.arcsinh(flux / (2. * bb_band)) + np.log(bb_band))
        return asinh_mag

    @hybridMag.expression
    def hybridMag(cls):
        if index is not None:
            # It needs to be index + 1 because Postgresql arrays are 1-indexed.
            flux = getattr(cls, flux_parameter)[index + 1]
        else:
            flux = getattr(cls, flux_parameter)

        flux *= 1e-9
        bb_band = bb[band]
        xx = flux / (2. * bb_band)
        asinh_mag = (-2.5 / func.log(10) *
                     (func.log(xx + func.sqrt(func.pow(xx, 2) + 1)) + func.log(bb_band)))
        return cast(asinh_mag, Float)

    return hybridMag


def HybridColour(parameter):

    @hybrid_method
    def colour(self, bandA, bandB):

        for band in [bandA, bandB]:
            columnName = parameter + '_' + band
            assert hasattr(self, columnName), \
                'cannot find column {0}'.format(columnName)

        bandA_param = getattr(self, parameter + '_' + bandA)
        bandB_param = getattr(self, parameter + '_' + bandB)

        return bandA_param - bandB_param

    @colour.expression
    def colour(cls, bandA, bandB):

        for band in [bandA, bandB]:
            columnName = parameter + '_' + band
            assert hasattr(cls, columnName), \
                'cannot find column {0}'.format(columnName)

        bandA_param = getattr(cls, parameter + '_' + bandA)
        bandB_param = getattr(cls, parameter + '_' + bandB)

        return bandA_param - bandB_param

    return colour


def HybridMethodToProperty(method, bandA, bandB):

    @hybrid_property
    def colour_property(self):
        return getattr(self, method)(bandA, bandB)

    return colour_property


def HybridPropertyArrayIndex(column, index):
    """Creates a hybrid property for a certain index of a column of type array."""

    @hybrid_property
    def array_band(self):
        return getattr(self, column)[index]

    return array_band


# Adds hybrid properties defining asinh magnitudes for each of the SDSS bands.
for ii, band in enumerate('ugriz'):
    propertyName = 'elpetro_mag_{0}'.format(band)
    setattr(NSA, propertyName, HybridMag('elpetro_flux', band, ii + 2))

# Creates attributes for apparent magnitudes.
setattr(NSA, 'elpetro_colour', HybridColour('elpetro_mag'))
for colour_a, colour_b in itertools.combinations('ugriz', 2):
    setattr(NSA, 'elpetro_mag_{0}_{1}'.format(colour_a, colour_b),
            HybridMethodToProperty('elpetro_colour', colour_a, colour_b))

# Expands elpetro_absmag into independent bands.
# TODO: see issue #211 about allowing direct queries on array. If we implement
# that we don't need to expand the bands.
for index, band in enumerate('FNugriz'):
    setattr(NSA, 'elpetro_absmag_{0}'.format(band),
            HybridPropertyArrayIndex('elpetro_absmag', index))

setattr(NSA, 'elpetro_absmag_colour', HybridColour('elpetro_absmag'))
for colour_a, colour_b in itertools.combinations('FNugriz', 2):
    setattr(NSA, 'elpetro_absmag_{0}_{1}'.format(colour_a, colour_b),
            HybridMethodToProperty('elpetro_absmag_colour', colour_a, colour_b))


# Add stellar mass hybrid attributes to NSA catalog
def logmass(parameter):

    @hybrid_property
    def mass(self):
        par = getattr(self, parameter)
        return math.log10(par) if par > 0. else 0.

    @mass.expression
    def mass(cls):
        par = getattr(cls, parameter)
        return cast(case([(par > 0., func.log(par)),
                          (par == 0., 0.)]), Float)

    return mass


setattr(NSA, 'elpetro_logmass', logmass('elpetro_mass'))
setattr(NSA, 'sersic_logmass', logmass('sersic_mass'))
#
configure_mappers()

sample_cache = RelationshipCache(MangaTarget.NSA_objects).\
    and_(RelationshipCache(MangaTarget.character)).\
    and_(RelationshipCache(CurrentCatalogue.catalogue))



# flake8: noqa

from . import cube, maps, modelcube, spaxel
from .cube import Cube
from .image import Image
from .maps import Maps
from .modelcube import ModelCube
from .plate import Plate
from .rss import RSS, RSSFiber
from .spaxel import Spaxel


# from .query import Query
# from .results import Results



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-10-11 17:51:43
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-29 17:23:15

from __future__ import print_function, division, absolute_import

import numpy as np
import marvin.tools
import matplotlib.pyplot as plt
from marvin import log, config
from .base import VACMixIn, VACTarget

class CallableDict(dict):
    """ Creates dictionary object with keys that can be called without using round brackets.
    Enables to execute functions inside the dictionary object implicitly.
    """
    
    def __getitem__(self, key):
        val = super().__getitem__(key)
        if callable(val):
            return val()
        return val

class FIREFLYVAC(VACMixIn):
    """Provides access to the MaNGA-FIREFLY VAC.

    VAC name: FIREFLY

    URL: https://www.sdss.org/dr17/manga/manga-data/manga-firefly-value-added-catalog/

    Description: Returns integrated and resolved stellar population parameters fitted by FIREFLY

    Authors: Justus Neumann, Jianhui Lian, Daniel Thomas, Claudia Maraston, and Lewis Hill

    """

    # Required parameters
    name = 'firefly'
    description = 'Returns stellar population parameters fitted by FIREFLY'
    version = {'DR15': 'v1_1_2', 'DR16': 'v1_1_2','DR17':'v3_1_1', 'MPL-11': 'v3_1_1'}
    display_name = 'Firefly'
    url = 'https://www.sdss.org/dr17/manga/manga-data/manga-firefly-value-added-catalog/'
    
    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps, marvin.tools.modelcube.ModelCube)

    # Required method
    def set_summary_file(self, release):
        ''' Sets the path to the Firefly summary file '''

        # define the variables to build a unique path to your VAC file
        # look up the MaNGA drpver from the release
        drpver, dapver = config.lookUpVersions(release)
		
        self.path_params = []
        if release in ['DR17', 'MPL-11']:
            
            self.path_params.append({'ver': self.version[release], 'drpver': drpver, 'models':'miles'})
            self.path_params.append({'ver': self.version[release], 'drpver': drpver, 'models':'mastar'})
			
            self.summary_file = [self.get_path('mangaffly', path_params=self.path_params[i]) for i in [0,1]]
            
            # specify a special data container for the general VAC tools
            self.data_container = dict(zip(['miles', 'mastar'], self.summary_file))
        else:
            self.path_params.append({'ver': self.version[release], 'drpver': drpver})

            # get_path returns False if the files do not exist locally
            self.summary_file = self.get_path('mangaffly', path_params=self.path_params[0])

    # Required method
    def get_target(self, parent_object):
        ''' Accesses VAC data for a specific target from a Marvin Tool object '''

        # get any parameters you need from the parent object
        plateifu = parent_object.plateifu
        imagesz = int(parent_object.header['NAXIS1'])

        if not isinstance(self.summary_file,list):   # for data releases <DR17
            # download the vac from the SAS if it does not already exist locally
            if not self.file_exists(self.summary_file):
                log.info('Warning: This file is ~6 GB.  It may take awhile to download')
                self.summary_file = self.download_vac('mangaffly', path_params=self.path_params)
                
            # create container for more complex return data.
            ffly = FFlyTarget(plateifu, vacfile=self.summary_file,imagesz=imagesz,release=parent_object.release)
            return ffly

        # for DR17 return dictionary; vacfile is only downloaded when key is selected
        ffly = CallableDict({
			"miles":lambda: self.prepare_container(plateifu, imagesz, 0, parent_object.release),
			"mastar":lambda: self.prepare_container(plateifu, imagesz, 1, parent_object.release)
            })

        return ffly
    
    def prepare_container(self,plateifu,imagesz,n,release):
        ''' Auxillary method for get_target(), used only for DR17.
        VAC access is transfered to this method to assure that vacfile is only downloaded when key is selected.'''
        
        if not self.file_exists(self.summary_file[n]):
            log.info('Warning: This file is ~6 GB.  It may take awhile to download')
            self.summary_file[n] = self.download_vac('mangaffly', path_params=self.path_params[n])
        container = FFlyTarget(plateifu, vacfile=self.summary_file[n], imagesz=imagesz,release=release)
        return container


class FFlyTarget(VACTarget):
    ''' A customized target class to also display Firefly 2-d maps

    This class handles data the Firefly summary file.  Row data from the summary
    file for the given target is returned via the `data` property. Specific Firefly
    parameters are available via the `stellar_pops` and `stellar_gradients` methods, respectively.
    2-d maps from the Firefly data can be produced via the `plot_map` method.

    TODO: 

    Parameters:
        targetid (str):
            The plateifu or mangaid designation
        vacfile (str):
            The path of the VAC summary file
        imagesz (int):
            The original array shape of the target cube

    Attributes:
        data:
            The target row data from the main VAC file
        targetid (str):
            The target identifier
    '''

    def __init__(self, targetid, vacfile, imagesz=None,release='DR17'):
        super(FFlyTarget, self).__init__(targetid, vacfile)
        self._image_sz = imagesz
        self._parameters = ['lw_age', 'mw_age', 'lw_z', 'mw_z']
        # select the index of the targetid from the main VAC extension
        self._idx = self._get_data(ext='GALAXY_INFO')['plateifu'] == targetid
        self.release = release
        

    def stellar_pops(self, parameter=None):
        ''' Returns the global stellar population properties

        Returns the global stellar population property within 1 Re for a given
        stellar population parameter.  If no parameter specified, returns the entire row.

        Parameters:
            parameter (str):
                The stellar population parameter to retrieve.  Can be one of ['lw_age', 'mw_age', 'lw_z', 'mw_z'].
        
        Returns:
            The data from the FIREFLY summary file for the target galaxy
        '''

        if parameter:
            assert parameter in self._parameters, 'parameter must be one of {0}'.format(
                self._parameters
            )

        if not self._indata:
            return "No FIREFLY result exists for {0}".format(self.targetid)

        if parameter:
            return self._get_data(ext='GLOBAL_PARAMETERS')[parameter + '_1re'][self._idx]
        else:
            return self._get_data(ext='GLOBAL_PARAMETERS')[self._idx]

    def stellar_gradients(self, parameter=None):
        ''' Returns the gradient of stellar population properties

        Returns the gradient of the stellar population property for a given
        stellar population parameter.  If no parameter specified, returns the entire row.

        Parameters:
            parameter (str):
                The stellar population parameter to retrieve.  Can be one of ['lw_age', 'mw_age', 'lw_z', 'mw_z'].
        
        Returns:
            The data from the FIREFLY summary file for the target galaxy
        '''

        if parameter:
            assert parameter in self._parameters, 'parameter must be one of {0}'.format(
                self._parameters
            )

        if not self._indata:
            return "No FIREFLY result exists for {0}".format(self.targetid)

        if parameter:
            return self._get_data(ext='GRADIENT_PARAMETERS')[parameter + '_gradient'][self._idx]
        else:
            return self._get_data(ext='GRADIENT_PARAMETERS')[self._idx]

    def _make_map(self, parameter=None):
        ''' Extract and create a 2d map '''

        params = self._parameters + [
            'e(b_v)',
            'stellar_mass',
            'surface_mass_density',
            'signal_noise',
        ]
        assert parameter in params, 'Parameter must be one of {0}'.format(
            params)

        # get the required arrays
        binid = self._get_data(ext='SPATIAL_BINID')[self._idx].squeeze(axis=0)
        bin1d = self._get_data(ext='SPATIAL_INFO')[self._idx, :, 0][0]
        try:
            prop = self._get_data(ext=parameter + '_voronoi')[self._idx, :, 0][0]
        except:
            prop = self._get_data(ext=parameter + '_voronoi')[self._idx, :][0]
        image_sz = self._image_sz

        # make a base map and reshape to a 1d array
        maps = (np.zeros((image_sz, image_sz)) -
                99).reshape(image_sz * image_sz)
        # find relevant elements in bin1d (anything not -9999)
        propinds = np.where(bin1d >= -1)
        # find relevant elements in binid (anything not -9999)
        inds = np.where(binid >= -1)
        # select the relevant elements from binid
        tmp = binid[inds]
        # find non-zero elements from the relevant elements
        newinds = np.where(tmp > -1)[0]
        # map the bin1d indices back to the original binid
        ai = bin1d[propinds].argsort()
        p = ai[np.searchsorted(bin1d[propinds], tmp[newinds], sorter=ai)]
        # replace map elements with the relevant prop parameter
        maps[newinds] = prop[p]
        # reshape the map array from 1d back to 2d
        maps = maps.reshape(image_sz, image_sz)

        return maps

    def plot_map(self, parameter=None, mask=None):
        ''' Plot map of stellar population properties
        
        Plots a 2d map of the specified FIREFLY stellar
        population parameter using Matplotlib.  Optionally mask
        the data when plotting using Numpy's Masked Array.  Default
        is to mask map values < -10.

        Parameters:
            parameter (str):
                The named of the VORONOI stellar pop. parameter
            mask (nd-array):
                A Numpy array of masked values to apply to the map

        Returns:
            The matplotlib axis image object

        '''

        if not self._indata:
            return "No FIREFLY result exists for {0}".format(self.targetid)

        # create the 2d map
        maps = self._make_map(parameter=parameter)
        
        # Correcting a minor bug in the DR15/DR16 VAC file: all S/N values are shifted by -9999. Adding +9999 to correct.
        # (maps elements with no data are set to -99 in the _make_map method) 
        if (parameter=='signal_noise') and (self.release!='DR17'):
            maps[maps!=-99]+=9999

        # only show the spaxels with non-empty values
        mask = (maps < -10) if not mask else mask
        masked_array = np.ma.array(maps, mask=mask)

        # plot the masked map
        fig, ax = plt.subplots()
        axim = ax.imshow(masked_array, interpolation='nearest',
                         cmap='RdYlBu_r', origin='lower')
        ax.set_xlabel('spaxel')
        ax.set_ylabel('spaxel')
        ax.set_title('Firefly {0}'.format(parameter.title()))

        # plot the colour bar
        cbar = fig.colorbar(axim, ax=ax, shrink=0.9)
        cbar.set_label(parameter.title(), fontsize=18, labelpad=20)
        cbar.ax.tick_params(labelsize=22)

        return axim

    def list_parameters(self):
        ''' List the parameters available for plotting '''

        params = self._parameters + [
            'e(b_v)',
            'stellar_mass',
            'surface_mass_density',
            'signal_noise',
        ]
        return params




# Licensed under a 3-clause BSD style license - see https://github.com/weaverba137/pydl/blob/master/licenses/LICENSE.rst
# -*- coding: utf-8 -*-
"""Python library for reading & writing yanny files.
yanny is an object-oriented interface to FTCL/yanny data files following
these specifications_.
The format of the returned object is similar to that returned by
``read_yanny()`` in the efftickle perl package (in the yannytools product).
Currently multidimensional arrays are only supported for type ``char``, and a
close reading of the specifications indicates that multidimensional arrays
were only ever intended to be supported for type ``char``.  So no
multidimensional arrays, sorry.
This file is a modified version of pydl's yanny reader. See original file at
https://github.com/weaverba137/pydl/blob/master/pydl/pydlutils/yanny.py
and the license at
https://github.com/weaverba137/pydl/blob/master/licenses/LICENSE.rst
.. _specifications: http://www.sdss3.org/dr8/software/par.php
"""
#
# Modules
#
from __future__ import print_function
import re
import os
import os.path
import datetime
import numpy
# commented out by parejkoj
import six
#from astropy.extern import six

from collections import OrderedDict

if six.PY3:
    long = int

#
# Classes
#
class yanny(OrderedDict):
    """An object interface to a yanny file.
    Create a yanny object using a yanny file, `filename`.  If the file exists,
    it is read, & the dict structure of the object will be basically the
    same as that returned by ``read_yanny()`` in the efftickle package.
    If the file does not exist, or if no filename is given, a blank
    structure is returned.  Other methods allow for subsequent writing
    to the file.
    Parameters
    ----------
    filename : str or file-like, optional
        The name of a yanny file or a file-like object representing a yanny file.
    np : bool, optional
        If ``True``, data in a yanny file will be converted into a NumPy record
        array. Default is ``False``.
    debug : bool, optional
        If ``True``, some simple debugging statements will be turned on. Default
        is ``False``.
    Attributes
    ----------
    np : bool
        If True, data in a yanny file will be converted into a NumPy record
        array.
    debug : bool
        If True, some simple debugging statements will be turned on.
    filename : str
        The name of a yanny parameter file.  If a file-like object was used
        to initialize the object, this will have the value 'in_memory.par'.
    _contents : str
        The complete contents of a yanny parameter file.
    _struct_type_caches : dict
        A dictionary of dictionaries, one dictionary for every structure
        definition in a yanny parameter file.  Contains the types of
        each column
    _struct_isarray_caches : dict
        A dictionary of dictionaries, one dictionary for every structure
        definition in a yanny parameter file.  Contains a boolean value
        for every column.
    _enum_cache : dict
        Initially ``None``, this attribute is initialized the first time
        the ``isenum()`` method is called.  The keyword is the name of the
        enum type, the value is a list of the possible values of that type.
    """
    #
    #
    #
    @staticmethod
    def get_token(string):
        """Removes the first 'word' from string.
        If the 'word' is enclosed in double quotes, it returns the
        contents of the double quotes. If the 'word' is enclosed in
        braces, it returns the contents of the braces, but does not
        attempt to split the array.  If the 'word' is the last word of the
        string, remainder is set equal to the empty string.  This is
        basically a wrapper on some convenient regular expressions.
        Parameters
        ----------
        string : str
            A string containing words.
        Returns
        -------
        get_token : tuple
            A tuple containing the first word and the remainder of the string.
        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.get_token("The quick brown fox")
        ('The', 'quick brown fox')
        """
        if string[0] == '"':
            (word, remainder) = re.search(r'^"([^"]*)"\s*(.*)',
                string).groups()
        elif string[0] == '{':
            (word, remainder) = re.search(r'^\{\s*([^}]*)\s*\}\s*(.*)',
                string).groups()
        else:
            try:
                (word, remainder) = re.split(r'\s+',string,1)
            except ValueError:
                (word, remainder) = (string, '')
        if remainder is None:
            remainder = ''
        return (word,remainder)
    #
    #
    #
    @staticmethod
    def protect(x):
        """Used to appropriately quote string that might contain whitespace.
        This method is mostly for internal use by the yanny object.
        Parameters
        ----------
        x : str
            The data to protect.
        Returns
        -------
        protect : str
            The data with white space protected by quotes.
        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.protect('This string contains whitespace.')
        '"This string contains whitespace."'
        >>> yanny.protect('This string contains a #hashtag.')
        '"This string contains a #hashtag."'
        """
        if isinstance(x,numpy.bytes_):
            s = x.decode()
        else:
            s = str(x)
        if len(s) == 0 or s.find('#') >= 0 or re.search(r'\s+',s) is not None:
            return '"' + s + '"'
        else:
            return s
    #
    #
    #
    @staticmethod
    def trailing_comment(line):
        """Identify a trailing comment and strip it.
        This routine works on the theory that a properly quoted comment mark
        will be surrounted by an odd number of double quotes, & we can
        skip to searching for the last one in the line.
        Parameters
        ----------
        line : str
            A line from a yanny file potentially containing trailing comments.
        Returns
        -------
        trailing_comment : str
            The line with any trailing comment and any residual white space
            trimmed off.
        Bugs
        ----
        This may fail in certain pathological cases, for example if a
        real trailing comment contains a single double-quote::
            # a 'pathological" trailing comment
        or if someone is over-enthusiastically commenting::
            # # # # # I like # characters.
        Examples
        --------
        >>> from pydl.pydlutils.yanny import yanny
        >>> yanny.trailing_comment('mystruct 1234 "#hashtag" # a comment.')
        'mystruct 1234 "#hashtag"'
        >>> yanny.trailing_comment('mystruct 1234 "#hashtag" # a "comment".')
        'mystruct 1234 "#hashtag"'
        """
        lastmark = line.rfind('#')
        if lastmark >= 0:
            #
            # Count the number of double quotes in the remainder of the line
            #
            if (len([c for c in line[lastmark:] if c == '"']) % 2) == 0:
                #
                # Even number of quotes
                #
                return line[0:lastmark].rstrip()
        return line
    #
    #
    #
    @staticmethod
    def dtype_to_struct(dt,structname='mystruct',enums=None):
        """Convert a NumPy dtype object describing a record array to
        a typedef struct statement.
        The second argument is the name of the structure.
        If any of the columns are enum types, enums must
        be a dictionary with the keys the column names, and the values
        are a tuple containing the name of the enum type as the first item
        and a tuple or list of possible values as the second item.
        Parameters
        ----------
        dt : numpy.dtype
            The dtype of a NumPy record array.
        structname : str, optional
            The name to give the structure in the yanny file.  Defaults to 'MYSTRUCT'.
        enums : dict, optional
            A dictionary containing enum information.  See details above.
        Returns
        -------
        dtype_to_struct : dict
            A dictionary suitable for setting the 'symbols' dictionary of a new
            yanny object.
        Examples
        --------
        """
        dtmap = {'i2':'short','i4':'int','i8':'long','f4':'float',
            'f8':'double'}
        returnenums = list()
        if enums is not None:
            for e in enums:
                lines = list()
                lines.append('typedef enum {')
                for n in enums[e][1]:
                    lines.append("    {0},".format(n))
                lines[-1] = lines[-1].strip(',')
                lines.append('}} {0};'.format(enums[e][0].upper()))
                returnenums.append("\n".join(lines))
                #lines.append('')
        lines = list()
        lines.append('typedef struct {')
        for c in dt.names:
            if dt[c].kind == 'V':
                t = dt[c].subdtype[0].str[1:]
                l = dt[c].subdtype[1][0]
            else:
                t = dt[c].str[1:]
                l = 0
            line = '    '
            if t[0].upper() in ['S', 'U']:
                if c in enums:
                    line += enums[c][0].upper()
                else:
                    line += 'char'
            else:
                line += dtmap[t]
            line += ' {0}'.format(c)
            if l > 0:
                line += "[{0:d}]".format(l)
            if t[0].upper() in ['S', 'U'] and c not in enums:
                line += "[{0:d}]".format(int(t[1:]))
            line += ';'
            lines.append(line)
        lines.append('}} {0};'.format(structname.upper()))
        return {structname.upper():list(dt.names),'enum':returnenums,'struct':["\n".join(lines)]}
    #
    #
    #
    def __init__(self,filename=None,string=None,np=False,debug=False):
        """Create a yanny object using a yanny file.
        @filename Contents of file will be read from provided filename.
        @string Object will be created from provided string.
        @np Convert numeric data into NumPy arrays?
        @debug Turn on simple debugging?
        """

        super(yanny, self).__init__()

        #
        # The symbol hash is inherited from the old read_yanny
        #
        self['symbols'] = OrderedDict()
        #
        # Create special attributes that contain the internal status of the object
        # this should prevent overlap with keywords in the data files
        #
        self.filename = ''
        self._contents = ''
        #
        # Since the re is expensive, cache the structure types keyed by the field.
        # Create a dictionary for each structure found.
        #
        self._struct_type_caches = OrderedDict()
        self._struct_isarray_caches = OrderedDict()
        self._enum_cache = None
        #
        # Optionally convert numeric data into NumPy arrays
        #
        self.np = np
        #
        # Turn on simple debugging
        #
        self.debug = debug

        if all([filename, string]):
            raise Exception("A yanny object can only be initialized with a filename OR string.")

        #
        # If the file exists, read it
        #
        if filename is not None:
            #
            # Handle file-like objects
            #
            if isinstance(filename, six.string_types):
                if os.access(filename,os.R_OK):
                    self.filename = filename
                    with open(filename,'r') as f:
                        self._contents = f.read()
            else:
                #
                # Assume file-like
                #
                self.filename = 'in_memory.par'
                self._contents = filename.read()
            self._parse()
        elif string is not None:
            #
            # Handle content passed in as string
            #
            if isinstance(string, six.string_types):
                self.filename = 'in_memory.par'
                self._contents = string
                self._parse()
            else:
                raise Exception("The value passed in for 'string' is not a string ({0})."\
                               .format(type(string)))
        return
    #
    #
    #
    def __str__(self):
        """Implement the ``str()`` function for yanny objects.
        Simply prints the current contents of the yanny file.
        """
        return self._contents
    #
    #
    #
    def __eq__(self,other):
        """Test two yanny objects for equality.
        Two yanny objects are assumed to be equal if their contents are equal.
        """
        if isinstance(other,yanny):
            return str(other) == str(self)
        return NotImplemented
    #
    #
    #
    def __ne__(self,other):
        """Test two yanny objects for inequality.
        Two yanny objects are assumed to be unequal if their contents are unequal.
        """
        if isinstance(other,yanny):
            return str(other) != str(self)
        return NotImplemented
    #
    #
    #
    def __bool__(self):
        """Give a yanny object a definite truth value.
        A yanny object is considered ``True`` if its contents are non-zero.
        """
        return len(self._contents) > 0

    # `__nonzero__` is needed for Python 2.
    # Python 3 uses `__bool__`.
    # http://stackoverflow.com/a/2233850/498873
    __nonzero__=__bool__

    #
    #
    #
    def type(self,structure,variable):
        """Returns the type of a variable defined in a structure.
        Returns ``None`` if the structure or the variable is undefined.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column whose type you want.
        Returns
        -------
        type : str
            The type of the variable.
        """
        if structure not in self:
            return None
        if variable not in self.columns(structure):
            return None
        #
        # Added code to cache values to speed up parsing large files.
        # 2009.05.11 / Demitri Muna, NYU
        # Find (or create) the cache for this structure.
        #
        try:
            cache = self._struct_type_caches[structure]
        except KeyError:
            self._struct_type_caches[structure] = OrderedDict()
            cache = self._struct_type_caches[structure] # cache for one struct type
        #
        # Lookup (or create) the value for this variable
        #
        try:
            var_type = cache[variable]
        except KeyError:
            if self.debug:
                print(variable)
            defl = [ x for x in self['symbols']['struct'] if x.find(structure.lower()) > 0 ]
            defu = [ x for x in self['symbols']['struct'] if x.find(structure.upper()) > 0 ]
            if len(defl) != 1 and len(defu) != 1:
                return None
            elif len(defl) == 1:
                definition = defl
            else:
                definition = defu
            typere = re.compile(r'(\S+)\s+{0}([\[<].*[\]>]|);'.format(variable))
            (typ,array) = typere.search(definition[0]).groups()
            var_type = typ + array.replace('<','[').replace('>',']')
            cache[variable] = var_type
        return var_type
    #
    #
    #
    def basetype(self,structure,variable):
        """Returns the bare type of a variable, stripping off any array information.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column whose type you want.
        Returns
        -------
        basetype : str
            The type of the variable, stripped of array information.
        """
        typ = self.type(structure,variable)
        if self.debug:
            print(variable, typ)
        try:
            return typ[0:typ.index('[')]
        except ValueError:
            return typ
    #
    #
    #
    def isarray(self,structure,variable):
        """Returns ``True`` if the variable is an array type.
        For character types, this means a two-dimensional array,
        *e.g.*: ``char[5][20]``.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column to check for array type.
        Returns
        -------
        isarray : bool
            ``True`` if the variable is an array.
        """
        try:
            cache = self._struct_isarray_caches[structure]
        except KeyError:
            self._struct_isarray_caches[structure] = OrderedDict()
            cache = self._struct_isarray_caches[structure]
        try:
            result = cache[variable]
        except KeyError:
            typ = self.type(structure,variable)
            character_array = re.compile(r'char[\[<]\d*[\]>][\[<]\d*[\]>]')
            if ((character_array.search(typ) is not None) or
                (typ.find('char') < 0 and (typ.find('[') >= 0
                or typ.find('<') >= 0))):
                cache[variable] = True
            else:
                cache[variable] = False
            result = cache[variable]
        return result
    #
    #
    #
    def isenum(self,structure,variable):
        """Returns true if a variable is an enum type.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column to check for enum type.
        Returns
        -------
        isenum : bool
            ``True`` if the variable is enum type.
        """
        if self._enum_cache is None:
            self._enum_cache = OrderedDict()
            if 'enum' in self['symbols']:
                for e in self['symbols']['enum']:
                    m = re.search(r'typedef\s+enum\s*\{([^}]+)\}\s*(\w+)\s*;',e).groups()
                    self._enum_cache[m[1]] = re.split(r',\s*',m[0].strip())
            else:
                return False
        return self.basetype(structure,variable) in self._enum_cache
    #
    #
    #
    def array_length(self,structure,variable):
        """Returns the length of an array type or 1 if the variable is not an array.
        For character types, this is the length of a two-dimensional
        array, *e.g.*, ``char[5][20]`` has length 5.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column to check for array length.
        Returns
        -------
        array_length : int
            The length of the array variable
        """
        if self.isarray(structure,variable):
            typ = self.type(structure,variable)
            return int(typ[typ.index('[')+1:typ.index(']')])
        else:
            return 1
    #
    #
    #
    def char_length(self,structure,variable):
        """Returns the length of a character field.
        *e.g.* ``char[5][20]`` is an array of 5 strings of length 20.
        Returns ``None`` if the variable is not a character type. If the
        length is not specified, *i.e.* ``char[]``, it returns the length of
        the largest string.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column to check for char length.
        Returns
        -------
        char_length : int or None
            The length of the char variable.
        """
        typ = self.type(structure,variable)
        if typ.find('char') < 0:
            return None
        try:
            return int(typ[typ.rfind('[')+1:typ.rfind(']')])
        except ValueError:
            if self.isarray(structure,variable):
                return max([max([len(x) for x in r]) for r in self[structure][variable]])
            else:
                return max([len(x) for x in self[structure][variable]])
    #
    #
    #
    def dtype(self,structure):
        """Returns a NumPy dtype object suitable for describing a table as a record array.
        Treats enums as string, which is what the IDL reader does.
        Parameters
        ----------
        structure : str
            The name of the structure.
        Returns
        -------
        dtype : numpy.dtype
            A dtype object suitable for describing the yanny structure as a record array.
        """

        str_dtype_letter = 'U' if six.PY3 else 'S'

        dt = list()
        dtmap = {'short':'i2', 'int':'i4', 'long':'i8', 'float':'f',
            'double':'d' }
        for c in self.columns(structure):
            typ = self.basetype(structure,c)
            if typ == 'char':
                d = "{0}{1:d}".format(str_dtype_letter, self.char_length(structure,c))
            elif self.isenum(structure,c):
                d = "{0}{1:d}".format(str_dtype_letter,
                                      max([len(x) for x in self._enum_cache[typ]]))
            else:
                d = dtmap[typ]
            if self.isarray(structure,c):
                dt.append((c,d,(self.array_length(structure,c),)))
            else:
                dt.append((c,d))
        dt = numpy.dtype(dt)
        return dt
    #
    #
    #
    def convert(self,structure,variable,value):
        """Converts value into the appropriate (Python) type.
        * ``short`` & ``int`` are converted to Python ``int``.
        * ``long`` is converted to Python ``long``.
        * ``float`` & ``double`` are converted to Python ``float``.
        * Other types are not altered.
        There may be further conversions into NumPy types, but this is the
        first stage.
        Parameters
        ----------
        structure : str
            The name of the structure that contains `variable`.
        variable : str
            The name of the column undergoing conversion.
        value : str
            The value contained in a particular row of `variable`.
        Returns
        -------
        convert : int, long, float or str
            `value` converted to a Python numerical type.
        """
        typ = self.basetype(structure,variable)
        if (typ == 'short' or typ == 'int'):
            if self.isarray(structure,variable):
                return [int(v) for v in value]
            else:
                return int(value)
        if typ == 'long':
            if self.isarray(structure,variable):
                return [long(v) for v in value]
            else:
                return long(value)
        if (typ == 'float' or typ == 'double'):
            if self.isarray(structure,variable):
                return [float(v) for v in value]
            else:
                return float(value)
        return value
    #
    #
    #
    def tables(self):
        """Returns a list of all the defined structures.
        This is just the list of keys of the object with the 'internal'
        keys removed.
        """
        foo = list()
        for k in self['symbols'].keys():
            if k not in ('struct','enum'):
                foo.append(k)
        return foo
    #
    #
    #
    def columns(self,table):
        """Returns an ordered list of column names associated with a particular table.
        The order is the same order as they are defined in the yanny file.
        Parameters
        ----------
        table : str
            The table whose columns are desired.
        Returns
        -------
        columns : list
            The list of column names.
        """
        foo = list()
        if table in self['symbols']:
            return self['symbols'][table]
        return foo
    #
    #
    #
    def size(self,table):
        """Returns the number of rows in a table.
        Parameters
        ----------
        table : str
            The table whose size desired.
        Returns
        -------
        size : int
            The number of rows in `table`.
        """
        foo = self.columns(table)
        return len(self[table][foo[0]])
    #
    #
    #
    def pairs(self):
        """Returns a list of keys to keyword/value pairs.
        Equivalent to doing ``self.keys()``, but with all the data tables &
        other control structures stripped out.
        """
        p = list()
        foo = self.tables()
        for k in self.keys():
            if k != 'symbols' and k not in foo:
                p.append(k)
        return p
    #
    #
    #
    def row(self,table,index):
        """Returns a list containing a single row from a specified table in column order
        If index is out of range, it returns an empty list.
        If the yanny object instance is set up for NumPy record arrays, then
        a single row can be obtained with::
            row0 = par['TABLE'][0]
        Parameters
        ----------
        table : str
            The table whose row is desired.
        index : int
            The number of the row to return.
        Returns
        -------
        row : list
            A row from `table`.
        """
        datarow = list()
        if table in self and index >= 0 and index < self.size(table):
            for c in self.columns(table):
                datarow.append(self[table][c][index])
        return datarow
    #
    #
    #
    def list_of_dicts(self, table):
        """Construct a list of dictionaries.
        Takes a table from the yanny object and constructs a list object
        containing one row per entry. Each item in the list is a dictionary
        keyed by the struct value names.
        If the yanny object instance is set up for NumPy record arrays, then
        the same functionality can be obtained with::
            foo = par['TABLE'][0]['column']
        Parameters
        ----------
        table : str
            The table to convert
        Returns
        -------
        list_of_dicts : list
            A list containing the rows of `table` converted to ``dict``.
        """
        return_list = list()
        d = OrderedDict()
        struct_fields = self.columns(table) # I'm assuming these are in order...
        for i in range(self.size(table)):
            one_row = self.row(table, i) # one row as a list
            j = 0
            for key in struct_fields:
                d[key] = one_row[j]
                j = j + 1
            return_list.append(OrderedDict(d)) # append a new dict (copy of d)
        return return_list
    #
    #
    #
    def new_dict_from_pairs(self):
        """Returns a new dictionary of keyword/value pairs.
        The new dictionary (*i.e.*, not a yanny object) contains the keys
        that ``self.pairs()`` returns. There are two reasons this is convenient:
        * the key 'symbols' that is part of the yanny object will not be present
        * a simple yanny file can be read with no further processing
        Example
        -------
        Read a yanny file and return only the pairs::
            >>> from os.path import dirname
            >>> from pydl.pydlutils.yanny import yanny
            >>> new_dict = yanny(dirname(__file__)+'/tests/t/test.par').new_dict_from_pairs()
            >>> new_dict['mjd']
            '54579'
            >>> new_dict['alpha']
            'beta gamma delta'
        added: Demitri Muna, NYU 2009-04-28
        """
        new_dictionary = OrderedDict()
        for key in self.pairs():
            new_dictionary[key] = self[key]
        return new_dictionary
    #
    #
    #
    def write(self,newfile=None,comments=None):
        """Write a yanny object to a file.
        This assumes that the filename used to create the object was not that
        of a pre-existing file.  If a file of the same name is detected,
        this method will *not* attempt to overwrite it, but will print a warning.
        This also assumes that the special 'symbols' key has been properly
        created.  This will not necessarily make the file very human-readable,
        especially if the data lines are long.  If the name of a new file is
        given, it will write to the new file (assuming it doesn't exist).
        If the writing is successful, the data in the object will be updated.
        Parameters
        ----------
        newfile : str, optional
            The name of the file to write.
        comments : str or list of str, optional
            Comments that will be placed at the head of the file.  If a
            single string is passed, it will be written out verbatim, so it
            had better contain '#' characters.  If a list of strings is
            passed, comment characters will be added and the strings
            will be joined together.
        """
        if newfile is None:
            if len(self.filename) > 0:
                newfile = self.filename
            else:
                raise ValueError("No filename specified!")
        if comments is None:
            basefile = os.path.basename(newfile)
            timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
            comments = "#\n# {0}\n#\n# Created by pydl.pydlutils.yanny.yanny\n#\n# {1}\n#\n".format(basefile,timestamp)
        else:
            if not isinstance(comments, str):
                comments = "\n".join(["# {0}".format(c) for c in comments]) + "\n"
        contents = comments
        #
        # Print any key/value pairs
        #
        for key in self.pairs():
            contents += "{0} {1}\n".format(key,self[key])
        #
        # Print out enum definitions
        #
        if len(self['symbols']['enum']) > 0:
            contents += "\n" + "\n\n".join(self['symbols']['enum']) + "\n"
        #
        # Print out structure definitions
        #
        if len(self['symbols']['struct']) > 0:
            contents += "\n" + "\n\n".join(self['symbols']['struct']) + "\n"
        contents += "\n"
        #
        # Print out the data tables
        #
        for sym in self.tables():
            columns = self.columns(sym)
            for k in range(self.size(sym)):
                line = list()
                line.append(sym)
                for col in columns:
                    if self.isarray(sym,col):
                        datum = '{' + ' '.join([self.protect(x) for x in self[sym][col][k]]) + '}'
                    else:
                        datum = self.protect(self[sym][col][k])
                    line.append(datum)
                contents += "{0}\n".format(' '.join(line))
        #
        # Actually write the data to file
        #
        if os.access(newfile,os.F_OK):
            print("{0} exists, aborting write!".format(newfile))
            print("For reference, here's what would have been written:")
            print(contents)
        else:
            with open(newfile,'w') as f:
                f.write(contents)
            self._contents = contents
            self.filename = newfile
            self._parse()
        return
    #
    #
    #
    def append(self,datatable):
        """Appends data to an existing FTCL/yanny file.
        Tries as much as possible to preserve the ordering & format of the
        original file.  The datatable should adhere to the format of the
        yanny object, but it is not necessary to reproduce the 'symbols'
        dictionary.  It will not try to append data to a file that does not
        exist.  If the append is successful, the data in the object will be updated.
        Parameters
        ----------
        datatable : dict
            The data to append.
        """
        if len(self.filename) == 0:
            raise ValueError("No filename is set for this object. Use the filename attribute to set the filename!")
        if not isinstance(OrderedDict, dict):
            raise ValueError("Data to append is not of the correct type. Use a dict!")
        timestamp = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
        contents = ''
        #
        # Print any key/value pairs
        #
        for key in datatable.keys():
            if key.upper() in self.tables() or key == 'symbols':
                continue
            contents += "{0} {1}\n".format(key, datatable[key])
        #
        # Print out the data tables
        #
        for sym in self.tables():
            if sym.lower() in datatable:
                datasym = sym.lower()
            else:
                datasym = sym
            if datasym in datatable:
                columns = self.columns(sym)
                for k in range(len(datatable[datasym][columns[0]])):
                    line = list()
                    line.append(sym)
                    for col in columns:
                        if self.isarray(sym,col):
                            datum = '{' + ' '.join([self.protect(x) for x in datatable[datasym][col][k]]) + '}'
                        else:
                            datum = self.protect(datatable[datasym][col][k])
                        line.append(datum)
                    contents += "{0}\n".format(' '.join(line))
        #
        # Actually write the data to file
        #
        if len(contents) > 0:
            contents = ("# Appended by yanny.py at {0}.\n".format(timestamp)) + contents
            if os.access(self.filename,os.W_OK):
                with open(self.filename,'a') as f:
                    f.write(contents)
                self._contents += contents
                self._parse()
            else:
                print("{0} does not exist, aborting append!".format(self.filename))
                print("For reference, here's what would have been written:")
                print(contents)
        else:
            print("Nothing to be appended!")
        return
    #
    #
    #
    def _parse(self):
        """Converts text into tables that users can use.
        This method is for use internally by the yanny object.  It is not
        meant to be called by users.
        Parsing proceeds in this order:
        #. Lines that end with a backslash character ``\\`` are reattached
           to following lines.
        #. Structure & enum definitions are identified, saved into the
           'symbols' dictionary & stripped from the contents.
        #. Structure definitions are interpreted.
        #. At this point, the remaining lines of the original file can only
           contain these things:
           * 'blank' lines, including lines that only contain comments
           * keyword/value pairs
           * structure rows
        #. The remaining lines are scanned sequentially.
           #. 'Blank' lines are identified & ignored.
           #. Whitespace & comments are stripped from non-blank lines.
           #. Empty double braces ``{{}}`` are converted into empty double
              quotes ``""``.
           #. If the first word on a line matches the name of a structure,
              the line is broken up into tokens & each token or set of tokens
              (for arrays) is converted to the appropriate Python type.
           #. If the first word on a line does not match the name of a
              structure, it must be a keyword, so this line is interpreted
              as a keyword/value pair.  No further processing is done to
              the value.
        #. At the conclusion of parsing, if ``self.np`` is ``True``, the
           structures are converted into NumPy record arrays.
        """
        #
        # there are five things we might find
        # 1. 'blank' lines including comments
        # 2. keyword/value pairs (which may have trailing comments)
        # 3. enumeration definitions
        # 4. structure definitions
        # 5. data
        #
        lines = self._contents
        #
        # Reattach lines ending with \
        #
        lines = re.sub(r'\\\s*\n',' ',lines)
        #
        # Find structure & enumeration definitions & strip them out
        #
        self['symbols']['struct'] = re.findall(r'typedef\s+struct\s*\{[^}]+\}\s*\w+\s*;',lines)
        self['symbols']['enum'] = re.findall(r'typedef\s+enum\s*\{[^}]+\}\s*\w+\s*;',lines)
        lines = re.sub(r'typedef\s+struct\s*\{[^}]+\}\s*\w+\s*;','',lines)
        lines = re.sub(r'typedef\s+enum\s*\{[^}]+\}\s*\w+\s*;','',lines)
        #
        # Interpret the structure definitions
        #
        typedefre = re.compile(r'typedef\s+struct\s*\{([^}]+)\}\s*(\w*)\s*;')
        for typedef in self['symbols']['struct']:
            typedefm = typedefre.search(typedef)
            (definition,name) = typedefm.groups()
            self[name.upper()] = OrderedDict()
            self['symbols'][name.upper()] = list()
            definitions = re.findall(r'\S+\s+\S+;',definition)
            for d in definitions:
                d = d.replace(';','')
                (datatype,column) = re.split(r'\s+',d)
                column = re.sub(r'[\[<].*[\]>]$','',column)
                self['symbols'][name.upper()].append(column)
                self[name.upper()][column] = list()
        comments = re.compile(r'^\s*#') # Remove lines containing only comments
        blanks = re.compile(r'^\s*$') # Remove lines containing only whitespace
        #
        # Remove trailing comments, but not if they are enclosed in quotes.
        #
        #trailing_comments = re.compile(r'\s*\#.*$')
        #trailing_comments = re.compile(r'\s*\#[^"]+$')
        double_braces = re.compile(r'\{\s*\{\s*\}\s*\}') # Double empty braces get replaced with empty quotes
        if len(lines) > 0:
            for line in lines.split('\n'):
                if self.debug:
                    print(line)
                if len(line) == 0:
                    continue
                if comments.search(line) is not None:
                    continue
                if blanks.search(line) is not None:
                    continue
                #
                # Remove leading & trailing blanks & comments
                #
                line = line.strip()
                line = self.trailing_comment(line)
                #line = trailing_comments.sub('',line)
                line = double_braces.sub('""',line)
                #
                # Now if the first word on the line does not match a
                # structure definition it is a keyword/value pair
                #
                (key, value) = self.get_token(line)
                uckey = key.upper()
                if uckey in self['symbols'].keys():
                    #
                    # Structure data
                    #
                    for column in self['symbols'][uckey]:
                        if len(value) > 0 and blanks.search(value) is None:
                            (data,value) = self.get_token(value)
                            if self.isarray(uckey,column):
                                #
                                # An array value
                                # if it's character data, it won't be
                                # delimited by {} unless it is a multidimensional
                                # string array.  It may or may not be delimited
                                # by double quotes
                                #
                                # Note, we're assuming here that the only
                                # multidimensional arrays are string arrays
                                #
                                arraydata = list()
                                while len(data) > 0:
                                    (token, data) = self.get_token(data)
                                    arraydata.append(token)
                                self[uckey][column].append(
                                    self.convert(uckey,column,arraydata))
                            else:
                                #
                                # A single value
                                #
                                self[uckey][column].append(
                                    self.convert(uckey,column,data))
                        else:
                            break
                else:
                    #
                    # Keyword/value pair
                    #
                    self[key] = value
        #
        # If self.np is True, convert tables into NumPy record arrays
        #
        if self.np:
            for t in self.tables():
                record = numpy.zeros((self.size(t),),dtype=self.dtype(t))
                for c in self.columns(t):
                    record[c] = self[t][c]
                self[t] = record
        return


def write_ndarray_to_yanny(filename, datatable, structname='mystruct',
                           enums=OrderedDict(),hdr=OrderedDict(), overwrite=False):
    """Converts a NumPy record array into a new FTCL/yanny file.
    Returns a new yanny object corresponding to the file.
    Parameters
    ----------
    filename : str
        The name of a parameter file.
    datatable : numpy.ndarray
        A NumPy record array containing data that can be copied into a
        yanny object.
    structname : str, optional
        The name to give the structure in the yanny file.  Defaults to
        'MYSTRUCT'.
    enums : dict, optional
        A dictionary containing enum information.  See details above.
    hdr : dict, optional
        A dictionary containing keyword/value pairs for the 'header' of the
        yanny file.
    Returns
    -------
    par : yanny.yanny
        The yanny object resulting from writing the file.
    Examples
    --------
    """

    par = yanny(filename, np=True, debug=False)

    par['symbols'] = par.dtype_to_struct(
        datatable.dtype, structname=structname, enums=enums)

    par[structname.upper()] = datatable

    for key in hdr:
        par[key] = hdr[key]

    if overwrite:
        os.remove(filename)

    par.write(filename)

    return par



#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-07-13
# @Filename: dapall.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-14 12:03:58

import distutils
import os

import astropy.io.fits

import marvin
from marvin.core.exceptions import MarvinError
from marvin.utils.general import get_dapall_path, map_dapall


__all__ = ['DAPallMixIn']


class DAPallMixIn(object):
    """A mixin that provides access to DAPall paremeters.

    Must be used in combination with `.MarvinToolsClass` and initialised
    before `~.DAPallMixIn.dapall` can be called.

    `DAPallMixIn` uses the `.MarvinToolsClass.data_origin` of the object to
    determine how to obtain the DAPall information. However, if the object
    contains a ``dapall`` attribute with the path to a DAPall file, that file
    will be used.

    """

    __min_dapall_version__ = distutils.version.StrictVersion('2.1.0')

    @property
    def dapall(self):
        """Returns the contents of the DAPall data for this target."""

        if (not self._dapver or
                distutils.version.StrictVersion(self._dapver) < self.__min_dapall_version__):
            raise MarvinError('DAPall is not available for versions before MPL-6.')

        if hasattr(self, '_dapall') and self._dapall is not None:
            return self._dapall

        if self.data_origin == 'file':
            try:
                dapall_data = self._get_dapall_from_file()
            except IOError:
                marvin.log.debug('cannot find DAPall file. Trying remote request.')
                dapall_data = self._get_from_api()
        elif self.data_origin == 'db':
            dapall_data = self._get_dapall_from_db()
        else:
            dapall_data = self._get_dapall_from_api()

        self._dapall = dapall_data

        return self._dapall

    def _get_dapall_from_file(self):
        """Uses DAPAll file to retrieve information."""

        daptype = self.bintype.name + '-' + self.template.name

        dapall_path = get_dapall_path(self._drpver, self._dapver)

        assert dapall_path is not None, 'cannot build DAPall file.'

        if not os.path.exists(dapall_path):
            raise MarvinError('cannot find DAPall file in the system.')

        dapall_hdu = astropy.io.fits.open(dapall_path)

        header = dapall_hdu[0].header
        dapall_table = dapall_hdu[-1].data

        dapall_row = dapall_table[(dapall_table['PLATEIFU'] == self.plateifu) &
                                  (dapall_table['DAPTYPE'] == daptype)]

        assert len(dapall_row) == 1, 'cannot find matching row in DAPall.'

        return map_dapall(header, dapall_row[0])

    def _get_dapall_from_db(self):
        """Uses the DB to retrieve the DAPAll data."""

        dapall_data = {}

        daptype = self.bintype.name + '-' + self.template.name

        mdb = marvin.marvindb

        if not mdb.isdbconnected:
            raise MarvinError('No DB connected')

        datadb = mdb.datadb
        dapdb = mdb.dapdb

        dapall_row = mdb.session.query(dapdb.DapAll).join(
            dapdb.File, datadb.PipelineInfo, datadb.PipelineVersion).filter(
                mdb.datadb.PipelineVersion.version == self._dapver,
                dapdb.DapAll.plateifu == self.plateifu,
                dapdb.DapAll.daptype == daptype).use_cache().first()

        if dapall_row is None:
            raise MarvinError('cannot find a DAPall match for this target in the DB.')

        for col in dapall_row.__table__.columns.keys():
            if col != 'pk' and '_pk' not in col:
                dapall_data[col] = getattr(dapall_row, col)

        return dapall_data

    def _get_dapall_from_api(self):
        """Uses the API to retrieve the DAPall data."""

        url = marvin.config.urlmap['api']['dapall']['url']

        url_full = url.format(name=self.plateifu,
                              bintype=self.bintype.name,
                              template=self.template.name)

        try:
            response = self._toolInteraction(url_full)
        except Exception as ee:
            raise MarvinError('found a problem while getting DAPall: {0}'.format(str(ee)))

        if response.results['error'] is not None:
            raise MarvinError('found a problem while getting DAPall: {}'
                              .format(str(response.results['error'])))

        data = response.getData()

        return data['dapall_data']



#!/usr/bin/env python
# encoding: utf-8
#
# maps.py
#
# Created by José Sánchez-Gallego on 25 Jun 2016.


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from flask_classful import route
from flask import jsonify

import marvin.api.base
import marvin.core.exceptions
import marvin.tools.maps
import marvin.utils.general


def _getMaps(name, **kwargs):
    """Returns a Maps object after parsing the name."""

    results = {}

    # Makes sure we don't use the wrong mode.
    kwargs.pop('mode', None)

    # Pop the release to remove a duplicate input to Maps
    release = kwargs.pop('release', None)

    # Parses name into either mangaid or plateifu
    try:
        idtype = marvin.utils.general.parseIdentifier(name)
    except Exception as ee:
        results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(ee))
        return None, results

    plateifu = None
    mangaid = None
    try:
        if idtype == 'plateifu':
            plateifu = name
        elif idtype == 'mangaid':
            mangaid = name
        else:
            raise marvin.core.exceptions.MarvinError(
                'invalid plateifu or mangaid: {0}'.format(idtype))

        maps = marvin.tools.maps.Maps(mangaid=mangaid, plateifu=plateifu,
                                      mode='local', release=release, **kwargs)
        results['status'] = 1
    except Exception as ee:
        maps = None
        results['error'] = 'Failed to retrieve maps {0}: {1}'.format(name, str(ee))

    return maps, results


class MapsView(marvin.api.base.BaseView):
    """Class describing API calls related to MaNGA Maps."""

    route_base = '/maps/'

    def index(self):
        '''Returns general maps info

        .. :quickref: Maps; Get general maps info

        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson string data: data message
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/maps/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": "this is a maps!"
           }

        '''
        self.results['status'] = 1
        self.results['data'] = 'this is a maps!'
        return jsonify(self.results)

    @route('/<name>/', defaults={'bintype': None, 'template': None},
           methods=['GET', 'POST'], endpoint='getMaps')
    @route('/<name>/<bintype>/', defaults={'template': None},
           methods=['GET', 'POST'], endpoint='getMaps')
    @route('/<name>/<bintype>/<template>/',
           methods=['GET', 'POST'], endpoint='getMaps')
    @marvin.api.base.arg_validate.check_args()
    def get(self, args, name, bintype, template):
        '''Returns the parameters needed to initialise a Maps remotely.

        .. :quickref: Maps; Get a maps given a name, bintype, and template

        :param name: The name of the maps as plate-ifu or mangaid
        :param bintype: The bintype associated with this maps.  If not defined, the default is used
        :param template: The template associated with this maps.  If not defined, the default is used
        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json string plateifu: id of maps
        :json string mangaid: mangaid of maps
        :json string header: the maps header as a string
        :json list shape: the maps shape [x, y]
        :json string wcs: the maps wcs_header as a string
        :json string bintype: the bintype of the maps
        :json string template: the template library of the maps
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/maps/8485-1901/SPX/GAU-MILESHC/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"plateifu": "8485-1901",
                       "mangaid": "1-209232",
                       "header": "XTENSION= 'IMAGE', NAXIS=3, .... END",
                       "wcs_header": "WCSAXES = 3 / Number of coordindate axes .... END",
                       "shape": [34, 34],
                       "bintype": "SPX",
                       "template": "GAU-MILESHC"
                       }
           }

        '''
        # Pop any args we don't want going into Maps
        args = self._pop_args(args, arglist='name')

        #kwargs = {'bintype': bintype, 'template_kin': template_kin}
        #maps, results = _getMaps(name, **kwargs)
        maps, results = _getMaps(name, **args)
        self.update_results(results)

        if maps is None:
            return jsonify(self.results)

        header = maps.header.tostring()
        wcs_header = maps.wcs.to_header_string()
        bintype = maps.bintype.name
        template = maps.template.name
        shape = maps._shape

        # Redefines plateifu and mangaid from the Maps
        mangaid = maps.mangaid
        plateifu = maps.plateifu

        self.results['data'] = {'mangaid': mangaid,
                                'plateifu': plateifu,
                                'header': header,
                                'wcs': wcs_header,
                                'bintype': bintype,
                                'template': template,
                                'shape': shape}

        return jsonify(self.results)

    @route('/<name>/<bintype>/<template>/map/<property_name>/',
           methods=['GET', 'POST'], endpoint='getmap', defaults={'channel': None})
    @route('/<name>/<bintype>/<template>/map/<property_name>/<channel>/',
           methods=['GET', 'POST'], endpoint='getmap')
    @marvin.api.base.arg_validate.check_args()
    def getMap(self, args, name, bintype, template, property_name, channel):
        """Returns data, ivar, mask, and unit for a given map.

        .. :quickref: Maps; Get map data given a name, bintype, template, property, and channel

        :param name: The name of the maps as plate-ifu or mangaid
        :param bintype: The bintype associated with this maps.  If not defined, the default is used
        :param template: The template associated with this maps.  If not defined, the default is used
        :param property_name: The property_name of the map to be extractred. E.g., `'emline_gflux'`.
        :param channel: If the ``property_name`` contains multiple channels, the channel to use, e.g., ``ha_6564'. Otherwise, ``None``
        :form release: the release of MaNGA data
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json list value: the data values of this map
        :json list ivar: the ivar values of this map
        :json list mask: the mask values of this map
        :json string unit: the unit on this channel for the given map property
        :json dict header: a dictionary of the header for this map
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/maps/8485-1901/SPX/GAU-MILESHC/map/emline_gflux/ha_6564/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"header": {"BITPIX": "-64", "C01": "OIId-3728", ...},
                       "unit": "1E-17 erg/s/cm^2/spaxel",
                       "value": [[0.0,0.0,0.0,...], [0,0,0,...], ... ],
                       "ivar": [[0.0,0.0,0.0,...], [0,0,0,...], ... ],
                       "mask": [[1073741843, 1073741843, 1073741843, ...], [1073741843, 1073741843, 1073741843, ...], ... ]
                      }
           }

        """

        # Pop any args we don't want going into Maps
        args = self._pop_args(args, arglist=['name', 'property_name', 'channel'])

        # kwargs = {'bintype': bintype, 'template_kin': template_kin}

        # Initialises the Maps object
        maps, results = _getMaps(name, **args)
        self.update_results(results)

        if maps is None:
            return jsonify(self.results)

        try:
            mmap = maps.getMap(property_name=str(property_name), channel=str(channel))
            self.results['data'] = {}
            self.results['data']['value'] = mmap.value.tolist()
            self.results['data']['ivar'] = mmap.ivar.tolist() if mmap.ivar is not None else None
            self.results['data']['mask'] = mmap.mask.tolist() if mmap.mask is not None else None
            self.results['data']['unit'] = mmap.unit.to_string()
        except Exception as ee:
            self.results['error'] = 'Failed to parse input name {0}: {1}'.format(name, str(ee))

        return jsonify(self.results)

    @route('/<name>/dapall', defaults={'bintype': None, 'template': None},
           methods=['GET', 'POST'], endpoint='dapall')
    @route('/<name>/<bintype>/dapall', defaults={'template': None},
           methods=['GET', 'POST'], endpoint='dapall')
    @route('/<name>/<bintype>/<template>/dapall',
           methods=['GET', 'POST'], endpoint='dapall')
    @marvin.api.base.arg_validate.check_args()
    def get_dapall_data(self, args, name, bintype, template):
        """Returns the DAPall data for a given mangaid or plateifu.

        .. :quickref: General; Returns the DAPall data for a given mangaid or plateifu.

        :param name: The name of the observation as mangaid or plateifu
        :param bintype: The bintype associated with this maps.  If not defined, the default is used
        :param template: The template associated with this maps.  If not defined, the default is used
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :json dict dapall_data: dict of the DAPall parameters
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/maps/8485-1901/SPX/GAU-MILESHC/dapall HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"dapall_data": {"plate": 7443,
                                       "ifudesign": 12701,
                                       ... }
                      }
           }

        """

        # Pop any args we don't want going into Maps
        args = self._pop_args(args, arglist='name')

        maps, results = _getMaps(name, **args)
        self.update_results(results)

        if maps is None:
            return jsonify(self.results)

        try:
            dapall = maps.dapall
            self.results['data'] = {'dapall_data': dapall}
        except marvin.core.exceptions.MarvinError as ee:
            self.results['error'] = str(ee)

        return jsonify(self.results)

    @route('/<name>/<bintype>/<template>/quantities/<x>/<y>/',
           methods=['GET', 'POST'],
           endpoint='getMapsQuantitiesSpaxel')
    @marvin.api.base.arg_validate.check_args()
    def getMapsQuantitiesSpaxel(self, args, name, bintype, template, x, y):
        """Returns a dictionary with all the quantities.

        .. :quickref: Maps; Returns a dictionary with all the quantities

        :param name: The name of the maps as plate-ifu or mangaid
        :param x: The x coordinate of the spaxel (origin is ``lower``)
        :param y: The y coordinate of the spaxel (origin is ``lower``)
        :form release: the release of MaNGA
        :resjson int status: status of response. 1 if good, -1 if bad.
        :resjson string error: error message, null if None
        :resjson json inconfig: json of incoming configuration
        :resjson json utahconfig: json of outcoming configuration
        :resjson string traceback: traceback of an error, null if None
        :resjson json data: dictionary of returned data
        :resheader Content-Type: application/json
        :statuscode 200: no error
        :statuscode 422: invalid input parameters

        **Example request**:

        .. sourcecode:: http

           GET /marvin/api/maps/8485-1901/quantities/10/12/ HTTP/1.1
           Host: api.sdss.org
           Accept: application/json, */*

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json
           {
              "status": 1,
              "error": null,
              "inconfig": {"release": "MPL-5"},
              "utahconfig": {"release": "MPL-5", "mode": "local"},
              "traceback": null,
              "data": {"emline_gflux_ha6564": {"value": 2.3, "ivar": ...},
                       "binid": ...}
              }
           }
        """

        # Pass the args in and get the cube
        args = self._pop_args(args, arglist=['name', 'x', 'y'])
        maps, res = _getMaps(name, **args)
        self.update_results(res)

        if maps:

            self.results['data'] = {}

            spaxel_quantities = maps._get_spaxel_quantities(x, y)

            for quant in spaxel_quantities:

                aprop = spaxel_quantities[quant]

                value = aprop.value
                ivar = aprop.ivar if aprop.ivar is not None else None
                mask = aprop.mask if aprop.mask is not None else None

                self.results['data'][quant] = {'value': value,
                                               'ivar': ivar,
                                               'mask': mask}

        return jsonify(self.results)



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-28 15:32:49
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-17 13:15:31

from __future__ import print_function, division, absolute_import
import os
import datetime
import redis


class Config(object):
    FLASK_APP = os.environ.get('FLASK_APP', 'marvin.web.uwsgi_conf_files.app')
    SECRET_KEY = os.environ.get('MARVIN_SECRET', 'secret-key')
    FLASK_SECRET = SECRET_KEY
    APP_DIR = os.path.abspath(os.path.dirname(__file__))  # This directory
    APP_BASE = os.environ.get('MARVIN_BASE', 'marvin')
    projroot = os.path.abspath(os.path.join(APP_DIR, os.pardir, os.pardir, os.pardir))
    PROJECT_ROOT = os.environ.get('MARVIN_DIR', projroot)
    BCRYPT_LOG_ROUNDS = 13
    ASSETS_DEBUG = False
    DEBUG_TB_ENABLED = False  # Disable Debug toolbar
    DEBUG_TB_INTERCEPT_REDIRECTS = False
    MAIL_SERVER = ''
    MAIL_PORT = 587
    MAIL_USE_SSL = False
    MAIL_USERNAME = ''
    MAIL_PASSWORD = ''
    MAIL_DEFAULT_SENDER = ''
    GOOGLE_ANALYTICS = ''
    LOG_SQL_QUERIES = True
    FEATURE_FLAGS_NEW = True
    UPLOAD_FOLDER = os.environ.get("MARVIN_DATA_DIR", '/tmp/')
    LIB_PATH = os.path.join(APP_DIR, 'lib')
    ALLOWED_EXTENSIONS = set(['txt', 'csv'])
    MAX_CONTENT_LENGTH = 16 * 1024 * 1024
    WTF_CSRF_ENABLED = True
    WTF_CSRF_SECRET_KEY = 'wtfsecretkey'
    USE_PROFILER = True  # Use the Flask Profiler Extension
    USE_SENTRY = False  # Turn off Sentry error logging
    FLASK_PROFILER = {
        "enabled": True,
        "storage": {
            "engine": "sqlite",
            "FILE": os.path.join(PROJECT_ROOT, 'flask_profiler.sql')
        },
        'endpointRoot': '{0}/profiler'.format(APP_BASE),
        "basicAuth": {
            "enabled": False
        },
        "ignore": [
            "/marvin/jsglue.js",
            "/marvin/static/.*",
            "/marvin/lib/.*",
            "/marvin/getgalidlist/",
            "/marvin/explore/webmap/"
        ]
    }
    # RATELIMIT_DEFAULT = '10/hour;100/day;2000 per year'
    RATELIMIT_STRATEGY = 'fixed-window-elastic-expiry'
    RATELIMIT_ENABLED = False

    # Flask-Session settings
    SESSION_TYPE = 'redis'
    SESSION_REDIS = redis.from_url(os.environ.get('SESSION_REDIS')
                                   ) if SESSION_TYPE == 'redis' and 'SESSION_REDIS' in os.environ else None
    SESSION_KEY_PREFIX = 'session:'

    # Flask-Caching settings
    CACHE_TYPE = 'redis'  # Can be "memcached", "redis", etc.
    CACHE_DEFAULT_TIMEOUT = 300
    CACHE_REDIS_URL = os.environ.get("SESSION_REDIS")
    CACHE_KEY_PREFIX = 'caching:'

    # Web/API Supported MPL/DR
    WEB_RELEASES = ['MPL-7', 'MPL-11']


class ProdConfig(Config):
    """Production configuration."""
    FLASK_ENV = os.environ.get('FLASK_ENV', 'production')
    ENV = 'prod'
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = 'postgresql://localhost/example'
    DEBUG_TB_ENABLED = False  # Disable Debug toolbar
    USE_X_SENDFILE = True
    USE_SENTRY = False
    SENTRY_DSN = os.environ.get('SENTRY_DSN', None)
    PERMANENT_SESSION_LIFETIME = datetime.timedelta(3600)
    JWT_ACCESS_TOKEN_EXPIRES = datetime.timedelta(300)
    REMEMBER_COOKIE_DOMAIN = '.sdss.org'


class DevConfig(Config):
    """Development configuration."""
    FLASK_ENV = os.environ.get('FLASK_ENV', 'development')
    ENV = 'dev'
    DEBUG = True
    DB_NAME = 'dev.db'
    # Put the db file in project root
    DB_PATH = os.path.join(Config.PROJECT_ROOT, DB_NAME)
    SQLALCHEMY_DATABASE_URI = 'sqlite:///{0}'.format(DB_PATH)
    DEBUG_TB_ENABLED = True
    ASSETS_DEBUG = True  # Don't bundle/minify static assets
    CACHE_TYPE = 'simple'  # Can be "memcached", "redis", etc.
    RATELIMIT_ENABLED = False


class DockerConfig(Config):
    """Docker configuration."""
    FLASK_ENV = os.environ.get('FLASK_ENV', 'production')
    ENV = 'prod'
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = 'postgresql://localhost/example'
    DEBUG_TB_ENABLED = False  # Disable Debug toolbar
    USE_X_SENDFILE = True
    USE_SENTRY = False
    SENTRY_DSN = os.environ.get('SENTRY_DSN', None)
    PERMANENT_SESSION_LIFETIME = datetime.timedelta(3600)
    JWT_ACCESS_TOKEN_EXPIRES = datetime.timedelta(300)
    REMEMBER_COOKIE_DOMAIN = '.sdss.org'
    SESSION_TYPE = 'filesystem'
    CACHE_TYPE = 'null'
    USE_PROFILER = False


class TestConfig(Config):
    FLASK_ENV = os.environ.get('FLASK_ENV', 'development')
    TESTING = True
    DEBUG = True
    SQLALCHEMY_DATABASE_URI = 'sqlite://'
    BCRYPT_LOG_ROUNDS = 1  # For faster tests
    WTF_CSRF_ENABLED = False  # Allows form testing
    PRESERVE_CONTEXT_ON_EXCEPTION = False
    USE_PROFILER = False  # Turn off the Flask Profiler extension
    RATELIMIT_ENABLED = False  # Turn off the Flask Rate Limiter
    #os.environ['PUBLIC_SERVER'] = 'True' # this breaks the debug server


class CustomConfig(object):
    ''' Project specific configuration.  Always gets appended to an above Config class '''
    os.environ['SAS_REDUX'] = 'sas/mangawork/manga/spectro/redux'
    os.environ['SAS_ANALYSIS'] = 'sas/mangawork/manga/spectro/analysis'
    os.environ['SAS_SANDBOX'] = 'sas/mangawork/manga/sandbox'
    release = os.environ.get('MARVIN_RELEASE', 'mangawork')
    os.environ['SAS_PREFIX'] = 'marvin' if release == 'mangawork' else 'dr17/marvin'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-08-17
# @Filename: structs.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-08 19:01:24


from __future__ import absolute_import, division, print_function

import gzip
import tempfile
from collections import OrderedDict
from contextlib import contextmanager

import six
from fuzzywuzzy import fuzz as fuzz_fuzz
from fuzzywuzzy import process as fuzz_proc


__ALL__ = ['FuzzyDict', 'Dotable', 'DotableCaseInsensitive', 'get_best_fuzzy',
           'FuzzyList', 'string_folding_wrapper', 'gunzip']


class Dotable(dict):
    """A custom dict class that allows dot access to nested dictionaries.

    Copied from http://hayd.github.io/2013/dotable-dictionaries/. Note that
    this allows you to use dots to get dictionary values, but not to set them.

    """

    def __getattr__(self, value):
        if '__' in value:
            return dict.__getattr__(self, value)
        else:
            return self.__getitem__(value)

    # def __init__(self, d):
    #     dict.__init__(self, ((k, self.parse(v)) for k, v in d.iteritems()))

    @classmethod
    def parse(cls, v):
        if isinstance(v, dict):
            return cls(v)
        elif isinstance(v, list):
            return [cls.parse(i) for i in v]
        else:
            return v


class DotableCaseInsensitive(Dotable):
    """Like dotable but access to attributes and keys is case insensitive."""

    def _match(self, list_of_keys, value):

        lower_values = [str(xx).lower() for xx in list_of_keys]
        if value.lower() in lower_values:
            return list_of_keys[lower_values.index(value.lower())]
        else:
            return False

    def __getattr__(self, value):
        if '__' in value:
            return super(DotableCaseInsensitive, self).__getattr__(value)
        return self.__getitem__(value)

    def __getitem__(self, value):
        key = self._match(list(self.keys()), value)
        if key is False:
            raise KeyError('{0} key or attribute not found'.format(value))
        return dict.__getitem__(self, key)


def get_best_fuzzy(value, choices, min_score=75, scorer=fuzz_fuzz.WRatio, return_score=False):
    """Returns the best match in a list of choices using fuzzywuzzy."""

    if not isinstance(value, six.string_types):
        raise ValueError('invalid value. Must be a string.')

    if len(value) < 3:
        raise ValueError('your fuzzy search value must be at least three characters long.')

    if len(choices) == 0:
        raise ValueError('choices cannot be an empty list.')

    # If the value contains _ivar or _mask this is probably and incorrect use
    # of the fuzzy feature. We raise an error.
    if '_ivar' in value:
        raise ValueError('_ivar not allowd in search value.')
    elif '_mask' in value:
        raise ValueError('_mask not allowd in search value.')

    bests = fuzz_proc.extractBests(value, choices, scorer=scorer, score_cutoff=min_score)

    if len(bests) == 0:
        best = None
    elif len(bests) == 1:
        best = bests[0]
    else:
        if bests[0][1] == bests[1][1]:
            best = None
        else:
            best = bests[0]

    if best is None:
        raise ValueError('cannot find a good match for {0!r}. '
                         'Your input value is too ambiguous.'.format(value))

    return best if return_score else best[0]


class FuzzyDict(OrderedDict):
    """A dotable dictionary that uses fuzzywuzzy to select the key."""

    def __getattr__(self, value):
        if '__' in value:
            return super(FuzzyDict, self).__getattr__(value)
        return self.__getitem__(value)

    def __getitem__(self, value):

        if not isinstance(value, six.string_types):
            return self.values()[value]

        if value in self.keys():
            return dict.__getitem__(self, value)

        best = get_best_fuzzy(value, self.keys())

        return dict.__getitem__(self, best)

    def __dir__(self):

        return list(self.keys())


class FuzzyList(list):
    """A list that uses fuzzywuzzy to select the item.

    Parameters:
        the_list (list):
            The list on which we will do fuzzy searching.
        use_fuzzy (function):
            A function that will be used to perform the fuzzy selection
    """

    def __init__(self, the_list, use_fuzzy=None):

        self.use_fuzzy = use_fuzzy if use_fuzzy else get_best_fuzzy

        list.__init__(self, the_list)

    def mapper(self, item):
        """The function that maps each item to the querable string."""

        return str(item)

    def __eq__(self, value):

        self_values = [self.mapper(item) for item in self]

        try:
            best = self.use_fuzzy(value, self_values)
        except ValueError:
            # Second pass, using underscores.
            best = self.use_fuzzy(value.replace(' ', '_'), self_values)

        return self[self_values.index(best)]

    def __contains__(self, value):

        if not isinstance(value, six.string_types):
            return super(FuzzyList, self).__contains__(value)

        try:
            self.__eq__(value)
            return True
        except ValueError:
            return False

    def __getitem__(self, value):

        if isinstance(value, six.string_types):
            return self == value
        else:
            return list.__getitem__(self, value)

    def __getattr__(self, value):

        self_values = [super(FuzzyList, self).__getattribute__('mapper')(item)
                       for item in self]

        if value in self_values:
            return self[value]

        return super(FuzzyList, self).__getattribute__(value)

    def __dir__(self):

        return [self.mapper(item) for item in self]


class OrderedDefaultDict(FuzzyDict):

    def __init__(self, default_factory=None, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        result = self[key] = self.default_factory()
        return result


class StringFolder(object):
    """
    Class that will fold strings. See 'fold_string'.
    This object may be safely deleted or go out of scope when
    strings have been folded.

    Credit: Ben Last
    http://dev.mobify.com/blog/sqlalchemy-memory-magic/

    """
    def __init__(self):
        self.unicode_map = {}

    def fold_string(self, s):
        """
        Given a string (or unicode) parameter s, return a string object
        that has the same value as s (and may be s). For all objects
        with a given value, the same object will be returned. For unicode
        objects that can be coerced to a string with the same value, a
        string object will be returned.
        If s is not a string or unicode object, it is returned unchanged.
        :param s: a string or unicode object.
        :return: a string or unicode object.
        """
        # If s is not a string or unicode object, return it unchanged
        if not isinstance(s, six.string_types):
            return s

        # If s is already a string, then str() has no effect.
        # If s is Unicode, try and encode as a string and use intern.
        # If s is Unicode and can't be encoded as a string, this try
        # will raise a UnicodeEncodeError.
        try:
            return six.moves.intern(str(s))
        except UnicodeEncodeError:
            # Fall through and handle s as Unicode
            pass

        # Look up the unicode value in the map and return
        # the object from the map. If there is no matching entry,
        # store this unicode object in the map and return it.
        t = self.unicode_map.get(s, None)
        if t is None:
            # Put s in the map
            t = self.unicode_map[s] = s
        return t


def string_folding_wrapper(results, keys=None):
    """
    This generator yields rows from the results as tuples,
    with all string values folded.

    Credit: Ben Last
    http://dev.mobify.com/blog/sqlalchemy-memory-magic/
    """
    # Get the list of keys so that we build tuples with all
    # the values in key order.

    if keys is None:
        try:
            keys = results.keys()
        except AttributeError as e:
            print('No keys are accessible.  Cannot fold strings!: {0}'.format(e))
            yield results

    folder = StringFolder()
    for row in results:
        yield tuple(
            folder.fold_string(row.__getattribute__(key))
            for key in keys
        )


@contextmanager
def gunzip(filename):
    """Context manager than gunzips a file temporarily."""

    temp_file = tempfile.NamedTemporaryFile(mode='wb')
    temp_file.file.write(gzip.GzipFile(filename).read())

    try:
        yield temp_file
    finally:
        temp_file.close()



# Licensed under a 3-clause BSD style license
"""
Marvin is a package intended to simply the access, exploration, and visualization of
the MaNGA dataset for SDSS-IV.  It provides a suite of Python tools, a web interface,
and a REST-like API, under tools/, web/, and api/, respectively.  Core functionality
of Marvin stems from Marvin's Brain.
"""

# isort:skip_file

import os
import re
import warnings
import contextlib
import six
from collections import OrderedDict
from astropy.wcs import FITSFixedWarning
from sdsstools import get_config, get_logger, get_package_version

NAME = 'marvin'

# Set the Marvin version
__version__ = get_package_version(path=__file__, package_name='sdss-marvin')

# Does this so that the implicit module definitions in extern can happen.
# time - 483 ms
from marvin.core.exceptions import MarvinUserWarning, MarvinError
from brain.utils.general.general import getDbMachine
from brain import bconfig
from brain.core.core import URLMapDict
from brain.core.exceptions import BrainError

# Loads config
curdir = os.path.dirname(os.path.abspath(__file__))
cfg_params = get_config(
    NAME, config_file=os.path.join(curdir, 'data/marvin.yml'))


# Defines log dir.
if 'MARVIN_LOGS_DIR' in os.environ:
    logFilePath = os.path.join(os.path.realpath(os.environ['MARVIN_LOGS_DIR']), 'marvin.log')
else:
    logFilePath = os.path.realpath(os.path.join(os.path.expanduser('~'), '.marvin', 'marvin.log'))

# Inits the log
log = get_logger(NAME)
log.start_file_logger(logFilePath)

warnings.simplefilter('once')
warnings.filterwarnings('ignore', 'Skipped unsupported reflection of expression-based index')
# warnings.filterwarnings('ignore', '(.)+size changed, may indicate binary incompatibility(.)+')
warnings.filterwarnings('ignore', category=FITSFixedWarning)

# This warning seems harmless (see https://github.com/astropy/astropy/issues/6025) so
# will ignore it for now.
warnings.filterwarnings('ignore', 'can\'t resolve package(.)+')

# Ignore DeprecationWarnings that are not Marvin's
warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('once', category=DeprecationWarning, module='marvin')

# Ignore numpy binary incompatibility runtime warning
warnings.filterwarnings('ignore', 'numpy.ufunc size changed')


# up to here - time: 1 second
class MarvinConfig(object):
    ''' Global Marvin Configuration

    The global configuration of Marvin.  Use the config object to globally set options for
    your Marvin session.

    Parameters:
        release (str):
            The release version of the MaNGA data you want to use.  Either MPL or DR.
        access (str):
            The type of access Marvin is allowed.  Either public or collab.
        download (bool):
            Set to turn on downloading of objects with sdss_access
        use_sentry (bool):
            Set to turn on/off the Sentry error logging.  Default is True.
        add_github_message (bool):
            Set to turn on/off the additional Github Issue message in MarvinErrors. Default is True.
        drpall (str):
            The location to your DRPall file, based on which release you have set.
        mode (str):
            The current mode of Marvin.  Either 'auto', 'remote', or 'local'. Default is 'auto'
        sasurl (str):
            The url of the Marvin API on the Utah Science Archive Server (SAS)
        urlmap (dict):
            A dictionary containing the API routing information used by Marvin
        xyorig (str):
            Globally set the origin point for all your spaxel selections.  Either 'center' or 'lower'.
            Default is 'center'
    '''
    def __init__(self):

        self._custom_config = None
        self._drpall = None
        self._dapall = None
        self._inapp = False

        self._urlmap = None
        self._xyorig = None

        self._release = None

        self.vermode = None
        self.download = False
        self.use_sentry = False
        self.add_github_message = True
        self._allowed_releases = {}

        # Allow DAP queries
        self._allow_DAP_queries = True

        # Allowed web releases; also see web.settings.Config.WEB_RELEASES
        self._web_releases = ['MPL-7', 'MPL-11']

        # perform some checks
        self._load_defaults()
        self._check_access()
        self._check_config()
        self._setDbConfig()

        # setup some paths
        self._plantTree()
        self._check_manga_dirs()
        self.setDefaultDrpAll()

    def _load_defaults(self):
        ''' Loads marvin configuration parameters

        Loads Marvin config parameters from the default marvin.yml file
        located in /data and merges with the user config parameters in
        ~/.marvin/marvin.yml

        '''
        self._custom_config = cfg_params

    def _checkPaths(self, name):
        ''' Check for the necessary path existence.

            This should only run if someone already has TREE_DIR installed
            but somehow does not have a SAS_BASE_DIR, MANGA_SPECTRO_REDUX, or
            MANGA_SPECTRO_ANALYSIS directory
        '''

        # set the access work path
        workpath = 'mangawork' if self.access == 'collab' else self.release.lower()

        name = name.upper()
        if name not in os.environ:
            if name == 'SAS_BASE_DIR':
                path_dir = os.path.expanduser('~/sas')
            elif name == 'MANGA_SPECTRO_REDUX':
                path_dir = os.path.join(os.path.abspath(os.environ['SAS_BASE_DIR']), '{0}/manga/spectro/redux'.format(workpath))
            elif name == 'MANGA_SPECTRO_ANALYSIS':
                path_dir = os.path.join(os.path.abspath(os.environ['SAS_BASE_DIR']), '{0}/manga/spectro/analysis'.format(workpath))

            if not os.path.exists(path_dir):
                warnings.warn('no {0}_DIR found. Creating it in {1}'.format(name, path_dir))
                os.makedirs(path_dir)
            os.environ[name] = path_dir

    @staticmethod
    def set_custom_path(name, path):
        ''' Set a temporary custom environment variable path

        Custom define a new environment variable in the current
        os session.  Puts it in your os.environ.  To permanently
        set a custom variable, use your .bashrc or .cshrc file.

        Parameters:
            name (str):
                The name of the environment variable
            path (str):
                The file path
        '''
        name = name.upper()
        os.environ[name] = path

    def _check_access(self):
        """Makes sure there is a valid netrc."""

        autocheck = self._custom_config.get('check_access', None)
        token = self._custom_config.get('use_token', None)

        if autocheck:
            try:
                valid_netrc = bconfig._check_netrc()
            except BrainError as e:
                pass
            else:
                # if it's valid, then auto set to collaboration
                if valid_netrc:
                    self.access = 'collab'

            # check for a valid login token in the marvin config
            if token and isinstance(token, six.string_types):
                self.token = token

    def _check_manga_dirs(self):
        """Check if $SAS_BASE_DIR and MANGA dirs are defined.
           If they are not, creates and defines them.
        """

        self._checkPaths('SAS_BASE_DIR')
        self._checkPaths('MANGA_SPECTRO_REDUX')
        self._checkPaths('MANGA_SPECTRO_ANALYSIS')

    def setDefaultDrpAll(self, drpver=None, dapver=None):
        """Tries to set the default location of drpall.

        Sets the drpall attribute to the location of your DRPall file, based on the
        drpver.  If drpver not set, it is extracted from the release attribute.  It sets the
        location based on the MANGA_SPECTRO_REDUX environment variable

        Parameters:
            drpver (str):
                The DRP version to set.  Defaults to the version corresponding to config.release.
        """

        if not drpver or not dapver:
            drpver, dapver = self.lookUpVersions(self.release)
        self.drpall = self._getDrpAllPath(drpver)
        self.dapall = self._getDapAllPath(drpver, dapver)

    def _get_default_path(self, name, drpver, dapver=None):
        ''' Return a default path to a summary file '''
        assert name in ['drpall', 'dapall'], 'name must be either drpall or dapall'
        envvar = 'MANGA_SPECTRO_REDUX' if name == 'drpall' else 'MANGA_SPECTRO_ANALYSIS'
        version = drpver if name == 'drpall' else dapver
        if name == 'drpall':
            path = os.path.join(str(drpver), 'drpall-{0}.fits'.format(version))
        elif name == 'dapall':
            path = os.path.join(str(drpver), str(dapver), 'dapall-{0}-{1}.fits'.format(drpver, dapver))

        if envvar in os.environ and version:
            return os.path.join(os.environ[envvar], path)
        else:
            raise MarvinError('Must have the {0} environment variable set'.format(envvar))

    def _getDrpAllPath(self, drpver):
        """Returns the default path for drpall, give a certain ``drpver``."""
        return self._get_default_path('drpall', drpver)

    def _getDapAllPath(self, drpver, dapver):
        """Returns the default path for dapall, give a certain ``drpver, dapver``."""
        return self._get_default_path('dapall', drpver, dapver=dapver)

############ Brain Config overrides ############
# These are configuration parameter defined in Brain.bconfig. We need
# to be able to modify them during run time, so we define properties and
# setters to do that from Marvin.config.

    @property
    def mode(self):
        return bconfig.mode

    @mode.setter
    def mode(self, value):
        bconfig.mode = value

    @property
    def sasurl(self):
        return bconfig.sasurl

    @sasurl.setter
    def sasurl(self, value):
        bconfig.sasurl = value

    @property
    def release(self):
        return self._release

    @release.setter
    def release(self, value):
        value = value.upper()
        if value not in self._allowed_releases:
            raise MarvinError('trying to set an invalid release version. Valid releases are: {0}'
                              .format(', '.join(self._allowed_releases)))

        # set the new release and possibly replant the tree
        with self._replant_tree(value) as val:
            self._release = val

        drpver, dapver = self.lookUpVersions(value)
        if 'MANGA_SPECTRO_REDUX' in os.environ:
            self.drpall = self._getDrpAllPath(drpver)

        if 'MANGA_SPECTRO_ANALYSIS' in os.environ:
            self.dapall = self._getDapAllPath(drpver, dapver)

    @property
    def access(self):
        return bconfig.access

    @access.setter
    def access(self, value):
        bconfig.access = value

        # update and recheck the releases
        self._check_config()

    @property
    def session_id(self):
        return bconfig.session_id

    @session_id.setter
    def session_id(self, value):
        bconfig.session_id = value

    @property
    def _traceback(self):
        return bconfig.traceback

    @_traceback.setter
    def _traceback(self, value):
        bconfig.traceback = value

    @property
    def compression(self):
        return bconfig.compression

    @compression.setter
    def compression(self, value):
        bconfig.compression = value

    @property
    def token(self):
        return bconfig.token

    @token.setter
    def token(self, value):
        bconfig.token = value

#################################################

    @property
    def urlmap(self):
        """Retrieves the URLMap the first time it is needed."""

        if self._urlmap is None or (isinstance(self._urlmap, dict) and len(self._urlmap) == 0):
            try:
                response = Interaction('/marvin/api/general/getroutemap', request_type='get', auth='netrc')
            except Exception as e:
                warnings.warn('Cannot retrieve URLMap. Remote functionality will not work: {0}'.format(e),
                              MarvinUserWarning)
                self.urlmap = URLMapDict()
            else:
                self.urlmap = response.getRouteMap()

        return self._urlmap

    @urlmap.setter
    def urlmap(self, value):
        """Manually sets the URLMap."""
        self._urlmap = value
        arg_validate.urlmap = self._urlmap

    @property
    def xyorig(self):
        if not self._xyorig:
            self._xyorig = 'center'

        return self._xyorig

    @xyorig.setter
    def xyorig(self, value):

        assert value.lower() in ['center', 'lower'], 'xyorig must be center or lower.'

        self._xyorig = value.lower()

    @property
    def drpall(self):
        return self._drpall

    @drpall.setter
    def drpall(self, value):
        if os.path.exists(value):
            self._drpall = value
        else:
            self._drpall = None
            warnings.warn('path {0} cannot be found. Setting drpall to None.'
                          .format(value), MarvinUserWarning)

    @property
    def dapall(self):
        return self._dapall

    @dapall.setter
    def dapall(self, value):
        if os.path.exists(value):
            self._dapall = value
        else:
            self._dapall = None
            warnings.warn('path {0} cannot be found. Setting dapall to None.'
                          .format(value), MarvinUserWarning)

    def _setDbConfig(self):
        ''' Set the db configuration '''
        self.db = getDbMachine()

    def _update_releases(self):
        ''' Update the allowed releases based on access '''

        # define release dictionaries
        mpldict = {'MPL-11': ('v3_1_1', '3.1.0'),
                   'MPL-10': ('v3_0_1', '3.0.1'),
                   'MPL-9': ('v2_7_1', '2.4.1'),
                   'MPL-8': ('v2_5_3', '2.3.0'),
                   'MPL-7': ('v2_4_3', '2.2.1'),
                   'MPL-6': ('v2_3_1', '2.1.3'),
                   'MPL-5': ('v2_0_1', '2.0.2'),
                   'MPL-4': ('v1_5_1', '1.1.1')}

        drdict = {'DR17': ('v3_1_1', '3.1.0'),
                  'DR15': ('v2_4_3', '2.2.1'),
                  'DR16': ('v2_4_3', '2.2.1')}

        # set the allowed releases based on access
        self._allowed_releases = {}
        if self.access == 'public':
            self._allowed_releases.update(drdict)
        elif self.access == 'collab':
            self._allowed_releases.update(drdict)
            self._allowed_releases.update(mpldict)

        # create and sort the final OrderedDict
        relsorted = sorted(self._allowed_releases.items(), key=lambda p: p[1][0], reverse=True)
        self._allowed_releases = OrderedDict(relsorted)

    def _get_latest_release(self, mpl_only=None, dr_only=None):
        ''' Get the latest release from allowed list '''

        if mpl_only:
            return max([r for r in list(self._allowed_releases) if 'MPL' in r], key=lambda t: int(t.split('-')[-1]))

        if dr_only:
            return max([r for r in list(self._allowed_releases) if 'DR' in r], key=lambda t: int(t[2:]))

        return list(self._allowed_releases)[0]

    def _check_config(self):
        ''' Check the release in the config '''

        # update the allowed releases
        self._update_releases()

        # check for a default release from the custom config
        default = 'default_release' in self._custom_config and self._custom_config['default_release']

        # Check for release version and if in allowed list
        latest = self._get_latest_release(mpl_only=self.access == 'collab')
        if not self.release:
            if default:
                self.setRelease(self._custom_config['default_release'])
            else:
                log.info('No release version set. Setting default to {0}'.format(latest))
                self.setRelease(latest)
        elif self.release and self.release not in self._allowed_releases:
            # this toggles to latest DR when switching to public
            warnings.warn('Release {0} is not in the allowed releases.  '
                          'Switching to {1}'.format(self.release, latest), MarvinUserWarning)

            self.setRelease(latest)

    def setRelease(self, version=None):
        """Set the release version.

        Parameters:
            version (str):
                The MPL/DR version to set, in form of MPL-X or DRXX.

        Example:
            >>> config.setRelease('MPL-4')
            >>> config.setRelease('DR13')

        """

        # if no version is set, pull the latest one by default
        if not version:
            version = self._get_latest_release()

        version = version.upper()
        self.release = version

    def setMPL(self, mplver):
        """As :func:`setRelease` but check that the version is an MPL."""

        assert 'MPL' in mplver.upper(), 'Must specify an MPL-X version!'
        self.setRelease(mplver)

    def setDR(self, drver):
        """As :func:`setRelease` but check that the version is a DR."""

        assert 'DR' in drver.upper(), 'Must specify a DRXX version!'
        self.setRelease(drver)

    def lookUpVersions(self, release=None):
        """Retrieve the DRP and DAP versions that make up a release version.

        Parameters:
            release (str or None):
                The release version. If ``None``, uses the currently set
                ``release`` value.

        Returns:
            drpver, dapver (tuple):
                A tuple of strings of the DRP and DAP versions according
                to the input MPL version

        """

        release = release or self.release

        try:
            drpver, dapver = self._allowed_releases[release]
        except KeyError:
            raise MarvinError('MPL/DR version {0} not found in lookup table. '
                              'No associated DRP/DAP versions. '
                              'Should they be added?  Check for typos.'.format(release))

        return drpver, dapver

    def lookUpRelease(self, drpver, public_only=None):
        """Retrieve the release version for a given DRP version

        Parameters:
            drpver (str):
                The DRP version to use
            public_only (bool):
                If True, uses only DR releases
        Returns:
            release (str):
                The release version according to the input DRP version
        """

        # Flip the mpldict
        verdict = {}
        for key, val in self._allowed_releases.items():
            if (public_only and 'MPL' in key):
                continue
            verdict[val[0]] = key

        try:
            release = verdict[drpver]
        except KeyError:
            raise MarvinError('DRP version {0} not found in lookup table. '
                              'No associated MPL version. Should one be added?  '
                              'Check for typos.'.format(drpver))

        return release

    def get_allowed_releases(self, public=None, min_release=None, 
                             web_releases=None):
        ''' Get a dictionary of supported MaNGA releases

        Set public flag to only include public releases. Set min_release to
        filer out all releases below this one.  Set web_releases to
        a filter out all releases included in the list.  

        Parameters:
            public (bool):
                If True, only return public DR releases
            min_release (str):
                Filter out all releases below this version
            web_releases (list):
                A list of all releases for the web/api

        Returns:
            dict: the supported MaNGA releases in Marvin
        '''
        allowed = self._allowed_releases.copy()

        # filter on only public DRs
        if public:
            allowed = {k: v for k, v in allowed.items() if 'DR' in k}

        # if minimum release set, filter on drpver >= min release version
        from packaging.version import parse
        if min_release:
            # adjust MPL-7 to DR15 alias if needed
            if min_release in ('MPL-7', 'MPL-11') and min_release not in allowed:
                min_release = 'DR15' if min_release == 'MPL-7' else 'DR17'
            min_drpver, __ = self.lookUpVersions(min_release)
            allowed = {k: v for k, v in allowed.items() if parse(
                v[0]) >= parse(min_drpver)}
            
        # if web releases set, filter them out
        if web_releases:
            allowed = {k: v for k, v in allowed.items() if k in web_releases or 'DR' in k}
        return allowed

    def switchSasUrl(self, sasmode='utah', ngrokid=None, port=5000, test=False,
                     base=None, public=None):
        ''' Switches the API SAS url config attribute

        Easily switch the sasurl configuration variable between
        utah and local.  utah sets it to the real API.  Local switches to
        use localhost.

        Parameters:
            sasmode ({'utah', 'local', 'mirror'}):
                the SAS mode to switch to.  Default is Utah
            ngrokid (str):
                The ngrok id to use when using a 'localhost' sas mode.
                This assumes localhost server is being broadcast by ngrok
            port (int):
                The port of your localhost server
            test (bool):
                If ``True``, sets the Utah sasurl to the test production, test/marvin
            base (str):
                The name of the marvin base.  Gets appended to main url.  Defaults to "marvin"
            public (bool):
                If ``True``, sets the API url to the public domain
        '''
        assert sasmode in ['utah', 'local', 'mirror'], 'SAS mode can only be utah, local, or mirror'
        base = base if base else os.environ.get('MARVIN_BASE', 'marvin')
        test_domain = 'https://lore.sdss.utah.edu/'
        if sasmode == 'local':
            if ngrokid:
                self.sasurl = 'http://{0}.ngrok.io/'.format(ngrokid)
            else:
                self.sasurl = 'http://localhost:{0}/'.format(port)
        elif sasmode == 'utah':
            # sort out the domain name
            if public:
                domain = test_domain if test else bconfig.public_api_url
                if test:
                    domain = '{0}{1}'.format(domain, 'public/')
                else:
                    domain = re.sub(r'(dr[0-9]{1,2})', self._release.lower(), domain)
            else:
                domain = test_domain if test else bconfig.collab_api_url

            url = os.path.join(domain)
            self.sasurl = url

            # get a new urlmap (need to do for vetting)
            self._urlmap = None
        elif sasmode == 'mirror':
            self.sasurl = bconfig.mirror_api_url

    def forceDbOff(self):
        ''' Force the database to be turned off '''
        config.db = None
        from marvin import marvindb
        if marvindb:
            marvindb.forceDbOff()

    def forceDbOn(self):
        ''' Force the database to be reconnected '''
        self._setDbConfig()
        from marvin import marvindb
        if marvindb:
            marvindb.forceDbOn(dbtype=self.db)

    def _plantTree(self):
        ''' Sets up the sdss tree product root '''

        tree_config = 'sdsswork' if self.access == 'collab' and 'MPL' in self.release else self.release.lower()

        try:
            from tree.tree import Tree
        except ImportError:
            self._tree = None
        else:
            self._tree = Tree(key='MANGA', config=tree_config)

    @contextlib.contextmanager
    def _replant_tree(self, value):
        ''' Replants the tree based on release/access toggling

        Context manager for use when setting a new release

        Replants the tree only when
        - A release is different than the previous value
        - Toggling to public access from sdsswork
        - Toggling to collab access from non-sdsswork
        - Toggling between DR releases
        - Toggling between DR and MPL releases

        Parameters:
            value (str):
                A new release
        '''

        # switch between public and collab access
        if hasattr(self, '_tree'):
            tocollab = self.access == 'collab' and self._tree.config_name != 'sdsswork'
            topublic = self.access == 'public' and self._tree.config_name == 'sdsswork'

        # switch trees based on release

        # check if release is different
        is_different = self._release and value != self._release

        # remove and return similar characters from value in self._release
        if self._release:
            similar = re.sub('[^{0}]'.format(self._release.replace('-', '\-')), '', value)
            # are we still a DR?
            stilldr = 'DR' in similar
            # are we still an MPL?
            stillmpl = 'MPL' in similar
            # have we changed releases?
            relchange = stilldr is False and stillmpl is False

        # yield the value (a release)
        yield value

        # check if the tree isn't set up or _release is None
        # this should only occur during the initial imports
        nullinit = not hasattr(self, '_tree') or self._release is None

        # set tree_config
        tree_config = 'sdsswork' if 'MPL' in value else value.lower() if 'DR' in value else None

        # replant the tree
        if is_different and not nullinit:
            if (relchange and self.access == 'collab') or stilldr or topublic or tocollab:
                self._tree.replant_tree(tree_config)

        # switch the API url depending on release
        if 'MPL' in value:
            self.switchSasUrl('utah')
        elif 'DR' in value:
            self.switchSasUrl('utah', public=True)

        # check to use the SAS mirror
        mirror = self._custom_config.get('use_mirror', None)
        if mirror:
            self.switchSasUrl('mirror', public='DR' in value)

    def login(self, refresh=None):
        ''' Login with netrc credentials to receive an API token

        Copy this token into the "use_token" parameter in your
        custom marvin.yml file to ensure preserve authentication
        across iPython user sessions.

        Parameters:
            refresh (bool):
                Set to True to refresh a login to receive a new token

        '''

        assert config.access == 'collab', 'You must have collaboration access to login.'

        # do nothing if token already generated
        if self.token and not refresh:
            return

        valid_netrc = bconfig._check_netrc()
        if valid_netrc:
            # get login info for api.sdss.org
            user, password = bconfig._read_netrc('api.sdss.org')
            data = {'username': user, 'password': password}

            # send token request
            url = self.urlmap['api']['login']['url']
            try:
                resp = Interaction(url, params=data, auth='netrc')
            except Exception as e:
                raise MarvinError('Error getting login token. {0}'.format(e))
            else:
                self.token = resp.results['access_token']


config = MarvinConfig()
# up to here - time: 1.6 seconds

# Inits the Database session and ModelClasses (time: 1.8 seconds)
marvindb = None
if config.db:
    from marvin.db.marvindb import MarvinDB
    marvindb = MarvinDB(dbtype=config.db, log=log, allowed_releases=config._allowed_releases.keys())

# Init MARVIN_DIR
marvindir = os.environ.get('MARVIN_DIR', None)
if not marvindir:
    moduledir = os.path.dirname(os.path.abspath(__file__))
    marvindir = moduledir.rsplit('/', 2)[0]
    os.environ['MARVIN_DIR'] = marvindir

from marvin.api.api import Interaction
from marvin.api.base import arg_validate

# Provide access to base submodules from the marvin namespace
from . import tools
from . import db
from . import utils



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-11-21 11:56:56
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-19 15:42:46

from __future__ import print_function, division, absolute_import
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils import statemachine
import traceback


def _indent(text, level=1):
    ''' Format Bintypes '''

    prefix = ' ' * (4 * level)

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if line.strip() else line)

    return ''.join(prefixed_lines())


def _format_datacubes(datacubes):
    ''' Format Datacubes table '''

    yield '.. list-table:: Datacubes'
    yield _indent(':widths: 15 50 50 10 10 20 20')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    yield _indent('  - Unit')
    yield _indent('  - Ivar')
    yield _indent('  - Mask')
    yield _indent('  - FITS')
    yield _indent('  - DB')

    for datacube in datacubes:
        dbcolumn = '{0}.{1}'.format(datacube.db_table, datacube.db_column())
        yield _indent('* - {0}'.format(datacube.name))
        yield _indent('  - {0}'.format(datacube.description))
        yield _indent('  - {0}'.format(datacube.unit.to_string()))
        yield _indent('  - {0}'.format(datacube.has_ivar()))
        yield _indent('  - {0}'.format(datacube.has_mask()))
        yield _indent('  - {0}'.format(datacube.fits_extension()))
        yield _indent('  - {0}'.format(dbcolumn))
    yield ''


def _format_rss(rss):
    ''' Format Rss table '''

    yield '.. list-table:: Rss'
    yield _indent(':widths: 15 50 50 10 10 20 20')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    yield _indent('  - Unit')
    yield _indent('  - Ivar')
    yield _indent('  - Mask')
    yield _indent('  - FITS')
    yield _indent('  - DB')

    for rs in rss:
        dbcolumn = '{0}.{1}'.format(rs.db_table, rs.db_column())
        yield _indent('* - {0}'.format(rs.name))
        yield _indent('  - {0}'.format(rs.description))
        yield _indent('  - {0}'.format(rs.unit.to_string()))
        yield _indent('  - {0}'.format(rs.has_ivar()))
        yield _indent('  - {0}'.format(rs.has_mask()))
        yield _indent('  - {0}'.format(rs.fits_extension()))
        yield _indent('  - {0}'.format(dbcolumn))
    yield ''


def _format_spectra(spectra):
    ''' Format Spectra '''

    yield '.. topic:: Spectra'
    yield '.. list-table:: Spectra'
    yield _indent(':widths: 15 100 20 20 20')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    yield _indent('  - Unit')
    yield _indent('  - FITS')
    yield _indent('  - DB')

    for spectrum in spectra:
        dbcolumn = '{0}.{1}'.format(spectrum.db_table, spectrum.db_column())
        yield _indent('* - {0}'.format(spectrum.name))
        yield _indent('  - {0}'.format(spectrum.description))
        yield _indent('  - {0}'.format(spectrum.unit.to_string()))
        yield _indent('  - {0}'.format(spectrum.fits_extension()))
        yield _indent('  - {0}'.format(dbcolumn))
    yield ''


def _format_bintypes(bintypes):
    ''' Format Bintypes '''

    yield '.. list-table:: Bintypes'
    yield _indent(':widths: 15 100 10')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    yield _indent('  - Binned')
    for bintype in bintypes:
        yield _indent('* - {0}'.format(bintype.name))
        yield _indent('  - {0}'.format(bintype.description))
        yield _indent('  - {0}'.format(bintype.binned))
    yield ''


def _format_templates(templates):
    ''' Format Templates '''

    yield '.. list-table:: Templates'
    yield _indent(':widths: 15 100')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    for template in templates:
        yield _indent('* - {0}'.format(template.name))
        yield _indent('  - {0}'.format(template.description))
    yield ''


def _format_models(models):
    ''' Format Models '''

    yield '.. list-table:: Models'
    yield _indent(':widths: 15 100 50 20 15 15')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Description')
    yield _indent('  - Unit')
    yield _indent('  - BinId')
    yield _indent('  - Ivar')
    yield _indent('  - Mask')

    for model in models:
        yield _indent('* - {0}'.format(model.name))
        yield _indent('  - {0}'.format(model.description))
        yield _indent('  - {0}'.format(model.unit))
        yield _indent('  - {0}'.format(model.binid.name))
        yield _indent('  - {0}'.format(model.has_ivar()))
        yield _indent('  - {0}'.format(model.has_mask()))

    yield ''


def _format_properties(properties):
    ''' Format Properties '''

    exts = properties.extensions

    yield '.. list-table:: Properties'
    yield _indent(':widths: 15 100 100 15 15 50 100')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Channels')
    yield _indent('  - Description')
    yield _indent('  - Ivar')
    yield _indent('  - Mask')
    yield _indent('  - FITS')
    yield _indent('  - DB')

    for prop in exts:
        yield _indent('* - {0}'.format(prop.name))
        if 'MultiChannelProperty' in str(prop.__class__):
            channels = ', '.join([c.name for c in prop.channels])
            dbcolumn = ', '.join(['{0}.{1}'.format(prop.db_table, c) for c in prop.db_columns()])
        else:
            channels = prop.channel
            dbcolumn = '{0}.{1}'.format(prop.db_table, prop.db_column())

        yield _indent('  - {0}'.format(channels))
        yield _indent('  - {0}'.format(prop.description))
        yield _indent('  - {0}'.format(prop.ivar))
        yield _indent('  - {0}'.format(prop.mask))
        yield _indent('  - {0}'.format(prop.fits_extension()))
        yield _indent('  - {0}'.format(dbcolumn))

    yield ''


def _format_parameters(parameters):
    ''' Format Query Parameters '''

    yield '.. topic:: Query Parameters'
    yield '.. list-table:: Query Parameters'
    yield _indent(':widths: 25 50 10 20 20 20 20')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Group')
    yield _indent('  - Full Name')
    yield _indent('  - Best')
    yield _indent('  - Name')
    yield _indent('  - DB Schema')
    yield _indent('  - DB Table')
    yield _indent('  - DB Column')

    for param in parameters:
        yield _indent('* - {0}'.format(param.group))
        yield _indent('  - {0}'.format(param.full))
        yield _indent('  - {0}'.format(param.best))
        yield _indent('  - {0}'.format(param.name))
        yield _indent('  - {0}'.format(param.db_schema))
        yield _indent('  - {0}'.format(param.db_table))
        yield _indent('  - {0}'.format(param.db_column))
    yield ''


def _format_schema(schema):
    ''' Format a maskbit schema '''

    schema_dict = schema.to_dict()
    indices = schema_dict['bit'].keys()

    yield '.. list-table:: Schema'
    yield _indent(':widths: 5 50 50')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Bit')
    yield _indent('  - Label')
    yield _indent('  - Description')

    for index in indices:
        yield _indent('* - {0}'.format(schema_dict['bit'][index]))
        yield _indent('  - {0}'.format(schema_dict['label'][index].strip()))
        yield _indent('  - {0}'.format(schema_dict['description'][index].strip()))
    yield ''


def _format_bitmasks(maskbit, bittype):
    ''' Format Maskbits '''

    for name, mask in maskbit.items():
        if bittype.lower() in name.lower():
            #yield '.. program:: {0}'.format(name)
            yield '{0}: {1}'.format(name, mask.description)
            yield ''
            for line in _format_schema(mask.schema):
                yield line


def _format_vacs(vacs, release):
    ''' Format a vac schema '''

    yield '.. list-table:: VACs'
    yield _indent(':widths: 20 10 50')
    yield _indent(':header-rows: 1')
    yield ''
    yield _indent('* - Name')
    yield _indent('  - Version')
    yield _indent('  - Description')

    for vac in vacs:
        yield _indent('* - {0}'.format(vac.name))
        yield _indent('  - {0}'.format(vac.version[release]))
        yield _indent('  - {0}'.format(vac.description))
    yield ''


def _format_command(name, command, **kwargs):
    """Format the output of `click.Command`."""

    # docstring
    # yield command.__doc__
    # yield ''

    # bintypes
    if 'bintypes' in kwargs:
        for line in _format_bintypes(command.bintypes):
            yield line

    # templates
    if 'templates' in kwargs:
        for line in _format_templates(command.templates):
            yield line

    # models
    if 'models' in kwargs:
        for line in _format_models(command.models):
            yield line

    # properties
    if 'properties' in kwargs:
        for line in _format_properties(command.properties):
            yield line

    # spectra
    if 'spectra' in kwargs:
        for line in _format_spectra(command.spectra):
            yield line

    # datacubes
    if 'datacubes' in kwargs:
        for line in _format_datacubes(command.datacubes):
            yield line

    # rss
    if 'rss' in kwargs:
        rssdm = kwargs.get('rssdm')
        for line in _format_rss(rssdm.rss):
            yield line

    # query parameters
    if 'parameters' in kwargs:
        for line in _format_parameters(command.parameters):
            yield line

    # bitmasks
    if 'bitmasks' in kwargs:
        for line in _format_bitmasks(command.bitmasks, kwargs.get('bittype', None)):
            yield line

    # vacs
    if 'vac' in kwargs:
        vac_release = kwargs.get('vac', None)
        if vac_release and vac_release in command:
            vacdm = command[vac_release]
            for line in _format_vacs(vacdm.vacs, vacdm.release):
                yield line


class DataModelDirective(rst.Directive):

    has_content = False
    required_arguments = 1
    option_spec = {
        'prog': directives.unchanged_required,
        'title': directives.unchanged,
        'subtitle': directives.unchanged,
        'description': directives.unchanged,
        'bintypes': directives.flag,
        'templates': directives.flag,
        'models': directives.flag,
        'properties': directives.flag,
        'datacubes': directives.flag,
        'rss': directives.flag,
        'spectra': directives.flag,
        'bitmasks': directives.flag,
        'parameters': directives.flag,
        'bittype': directives.unchanged,
        'vac': directives.unchanged,
    }

    def _load_module(self, module_path):
        """Load the module."""

        # __import__ will fail on unicode,
        # so we ensure module path is a string here.
        module_path = str(module_path)

        try:
            module_name, attr_name = module_path.split(':', 1)
        except ValueError:  # noqa
            raise self.error('"{0}" is not of format "module:parser"'.format(module_path))

        try:
            mod = __import__(module_name, globals(), locals(), [attr_name])
        except (Exception, SystemExit) as exc:  # noqa
            err_msg = 'Failed to import "{0}" from "{1}". '.format(attr_name, module_name)
            if isinstance(exc, SystemExit):
                err_msg += 'The module appeared to call sys.exit()'
            else:
                err_msg += 'The following exception was raised:\n{0}'.format(traceback.format_exc())

            raise self.error(err_msg)

        if not hasattr(mod, attr_name):
            raise self.error('Module "{0}" has no attribute "{1}"'.format(module_name, attr_name))

        return getattr(mod, attr_name)

    def _generate_nodes(self, name, command, parent=None, options={}):
        """Generate the relevant Sphinx nodes.
        Format a `click.Group` or `click.Command`.
        :param name: Name of command, as used on the command line
        :param command: Instance of `click.Group` or `click.Command`
        :param parent: Instance of `click.Context`, or None
        :param show_nested: Whether subcommands should be included in output
        :returns: A list of nested docutil nodes
        """

        # Title
        source_name = name
        content = [nodes.title(text=name)]
        subtitle = self.options.get('subtitle', None)
        description = self.options.get('description', None)

        if subtitle:
            content.append(nodes.subtitle(text=subtitle))
        if description:
            content.append(nodes.paragraph(text=description))

        section = nodes.section(
            '',
            *content,
            ids=[nodes.make_id(source_name)],
            names=[nodes.fully_normalize_name(source_name)])

        # Summary

        result = statemachine.ViewList()
        lines = _format_command(name, command, **options)
        for line in lines:
            result.append(line, source_name)
        self.state.nested_parse(result, 0, section)

        return [section]

    def run(self):
        self.env = self.state.document.settings.env

        # load the designated class object from the module file
        command = self._load_module(self.arguments[0])

        # do something special to access the RSS datamodel
        if 'rss' in self.options:
            rssarg = self.arguments[0].split(':')[0] + ':datamodel_rss'
            rssdms = self._load_module(rssarg)
            rssdm = rssdms[command.release]
            self.options['rssdm'] = rssdm

        if 'prog' in self.options:
            prog_name = self.options.get('prog')
        else:
            raise self.error(':prog: must be specified')

        return self._generate_nodes(prog_name, command, None, options=self.options)


def setup(app):
    app.add_directive('datamodel', DataModelDirective)






# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: MPL11.py
# Project: dap
# Author: Brian Cherinka
# Created: Friday, 26th February 2021 9:23:41 am
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2021 Brian Cherinka
# Last Modified: Friday, 26th February 2021 9:23:42 am
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import copy

from .base import DAPDataModel, Template
from .MPL10 import MILESHC_MASTARHC2, SPX, HYB10, VOR10, MPL10_maps, MPL10_models, binid_properties
from marvin.utils.datamodel.maskbit import get_maskbits

# update Template for MPL-11
MILESHC_MASTARSSP = Template('MILESHC-MASTARSSP',
                             description=('Stellar kinematic templates from the MILES library.  '
                                          'Stellar continuum template derived from a subset of the'
                                          'MaStar Simple Stellar Population models; '
                                          'used during emission-line fits.'))


MPL11_maps = copy.deepcopy(MPL10_maps)
MPL11_models = copy.deepcopy(MPL10_models)


# MPL-11 DapDataModel goes here
MPL11 = DAPDataModel('3.1.0', aliases=['MPL-11', 'MPL11', 'DR17'],
                     bintypes=[SPX, HYB10, VOR10],
                     templates=[MILESHC_MASTARHC2, MILESHC_MASTARSSP],
                     excluded_daptypes=['SPX-MILESHC-MASTARHC2', 'VOR10-MILESHC-MASTARHC2'],
                     properties=MPL11_maps,
                     models=MPL11_models,
                     bitmasks=get_maskbits('MPL-11'),
                     default_bintype='HYB10',
                     default_template='MILESHC-MASTARSSP',
                     property_table='SpaxelProp11',
                     default_binid=copy.deepcopy(binid_properties[0]),
                     default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                     qual_flag='DAPQUAL')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2018-07-14
# @Filename: aperture.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-07-27 13:02:19

import astropy.coordinates
import astropy.units
import numpy


try:
    import photutils
except ImportError:
    photutils = None


__all__ = ['GetApertureMixIn', 'MarvinAperture']


class MarvinAperture(photutils.Aperture if photutils else object):
    """Extends `photutils.Aperture` allowing to extract spaxels in the aperture.

    This class is not intended for general use and it is dynamically set as
    the base for a `~photutils.Aperture` instance.

    """

    @property
    def parent(self):
        """Returns or sets the parent object."""

        if not hasattr(self, '_parent'):
            return None

        return self._parent

    @parent.setter
    def parent(self, value):

        self._parent = value

    @property
    def mask(self):
        """Returns the fractional overlap mask.

        Equivalent to using `photutils.PixelAperture.to_mask` followed
        by `photutils.ApertureMask.to_image` using the shape of the parent
        object. Combines all the apertures in a single mask.

        """

        assert self.parent is not None, 'no parent set'

        if isinstance(self, photutils.SkyAperture):
            aperture = self.to_pixel(self.parent.wcs)
        else:
            aperture = self

        mask = numpy.zeros(self.parent._shape)

        for ap in aperture.to_mask(method='exact'):
            mask += ap.to_image(shape=(self.parent._shape))

        return mask

    def getSpaxels(self, threshold=0.5, lazy=True, mask=None, **kwargs):
        """Returns the spaxels that fall within the aperture.

        Parameters
        ----------
        threshold : float
            The minimum fractional overlap between the spaxel and the aperture
            grid for the spaxel to be returned.
        lazy : bool
            Whether the returned spaxels must be fully loaded or lazily
            instantiated.
        mask : numpy.ndarray
            A mask that defines the fractional pixel overlap with the
            apertures. If ``None``, the mask returned by `.MarvinAperture.mask`
            will be used.
        kwargs : dict
            Additional parameters to be passed to the parent ``getSpaxel``
            method. Can be used to define what information is loaded
            in the spaxels.

        """

        assert threshold > 0 and threshold <= 1, 'invalid threshold value'

        if mask is None:
            mask = self.mask
        else:
            assert mask.shape == self.parent._shape, 'invalid mask shape'

        spaxel_coords = numpy.where(mask >= threshold)

        if len(spaxel_coords[0]) == 0:
            return []

        return self.parent.getSpaxel(x=spaxel_coords[1], y=spaxel_coords[0],
                                     xyorig='lower', lazy=lazy, **kwargs)


class GetApertureMixIn(object):

    def getAperture(self, coords, aperture_params, aperture_type='circular',
                    coord_type='pixel'):
        """Defines an aperture.

        This method is mostly a wrapper around the aperture classes defined in
        `photutils <http://photutils.readthedocs.io>`_. It allows to

        Parameters
        ----------
        coords : tuple or `~numpy.ndarray`
            Either a 2-element tuple ``(x, y)`` or ``(ra, dec)`` to define the
            centre of a single aperture, or a list of N tuples or a Nx2 array
            to define N apertures.
        aperture_params : tuple
            A tuple with the parameters of the aperture.

            * For ``aperture_type='rectangular'``:

              * If ``coord_type='pixel'``, a tuple ``(w, h, theta)`` where
                ``w`` is the full width of the aperture (for ``theta=0`` the
                width side is along the ``x`` axis); ``h`` is the full height
                of the aperture (for ``theta=0`` the height side is along the
                ``y`` axis); and ``theta`` is the rotation angle in radians of
                the width (``w``) side from the positive ``x`` axis (the
                rotation angle increases counterclockwise).
              * If ``coord_type='sky'``, same format but ``w`` and ``h`` must
                be in arcsec and ``theta`` is the position angle (in degrees)
                of the width side. The position angle increases
                counterclockwise, from North (PA=0) towards East.

            * For ``aperture_type='circular'``:

              * The radius ``r`` in units of pixels or arcsec depending on the
                value of ``coord_type``. Can be a tuple or a float.

            * For ``aperture_type='elliptical'``:

              * If ``coord_type='pixel'``, a tuple ``(a, b, theta)`` where
                ``a`` and ``b`` are the semi-major and semi-minor axes of the
                ellipse, respectively, and ``theta`` is the rotation angle in
                radians of the semi-major axis from the positive x axis (the
                rotation angle increases counterclockwise). If
                ``coord_type='sky'``, ``a`` and ``b`` must be in arcsec, and
                ``theta`` is the position angle (in degrees) of the semi-major
                axis. The position angle increases counterclockwise, from North
                (PA=0) towards East.

        aperture_type : {'rectangular', 'circular', 'elliptical'}
            The type of aperture to define.
        coord_type : {'pixel', 'sky'}
            Determines whether the coordinates and aperture parameters refer
            to to the frame of the image or to sky coordinates. The conversion
            between the image and sky frames is determined using the WCS
            headers from the image.

        Returns
        -------
        marvin_aperture : MarvinAperture object
            A `.MarvinAperture` instance with the definition of the aperture,
            which can be used to extract the associated spaxels or to return
            the mask.

        Examples
        --------

        A single circular aperture with a radius of 3 pixels can created as ::

            >>> cube = marvin.tools.Cube('8485-1901')
            >>> aperture = cube.getAperture((17, 15), 3)

        ``aperture`` can then be used to return all spaxels that have a
        fractional overlap with the aperture of more than 0.6 pixels ::

            >>> spaxels = aperture.getSpaxels(threshold=0.6, lazy=True)
            >>> spaxels[0]
            <Marvin Spaxel (x=15, y=13, loaded=False)

        Apertures can also be defined from sky coordinates. To define two
        elliptical apertures with semi-axes 3 and 1 arcsec and rotated
        30 degrees we do ::

            >>> ap_ell = cube.getAperture([(232.546173, 48.6892288), (232.544069, 48.6906177)],
                                          (3, 1, 30), aperture_type='elliptical')
            >>> ap_ell
            <MarvinAperture([[232.546173 ,  48.6892288],
                 [232.544069 ,  48.6906177]], a=3.0, b=1.0, theta=30.0)>

        """

        if photutils is None:
            raise ImportError('this feature requires photutils. Install it by '
                              'doing pip install photutils.')

        assert coord_type in ['pixel', 'sky'], 'invalid coord_type'

        if isinstance(coords, astropy.coordinates.SkyCoord):
            coord_type = 'sky'
        else:
            coords = numpy.atleast_2d(coords)
            if coord_type == 'sky':
                coords = astropy.coordinates.SkyCoord(coords, unit='deg')

        aperture_params = numpy.atleast_1d(aperture_params).tolist()

        if aperture_type == 'circular':
            if coord_type == 'pixel':
                ApertureClass = photutils.CircularAperture
            else:
                ApertureClass = photutils.SkyCircularAperture
        elif aperture_type == 'elliptical':
            if coord_type == 'pixel':
                ApertureClass = photutils.EllipticalAperture
            else:
                ApertureClass = photutils.SkyEllipticalAperture
        elif aperture_type == 'rectangular':
            if coord_type == 'pixel':
                ApertureClass = photutils.RectangularAperture
            else:
                ApertureClass = photutils.SkyRectangularAperture
        else:
            raise ValueError('invalid aperture_type')

        # If on-sky, converts aperture parameters to quantities
        if coord_type == 'sky':

            if aperture_type == 'circular':
                n_params = 1
            else:
                n_params = 3

            assert len(aperture_params) == n_params, 'invalid number of parameters'

            units = [astropy.units.arcsec, astropy.units.arcsec, astropy.units.deg]

            for ii in range(n_params):
                if not isinstance(aperture_params[ii], astropy.units.Quantity):
                    aperture_params[ii] *= units[ii]

        aperture = ApertureClass(coords, *aperture_params)

        # Overrides the aperture class so that it inherits from MarvinAperture and
        # can gain the methods we defined there. Sets the parent to self.
        aperture.__class__ = type('MarvinAperture', (ApertureClass, MarvinAperture), {})
        aperture.parent = self

        return aperture



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-09-20 10:31:11
# @Last modified by:   Brian Cherinka
# @Last modified time: 2018-08-04 13:22:38

from __future__ import absolute_import, division, print_function

import copy
from collections import OrderedDict
import six
from marvin import config


class MetaDataModel(type):
    ''' MetaClass to construct a new DataModelList class '''
    def __new__(cls, name, parents, dict):
        if 'base' in dict:
            item = list(dict['base'].items())[0]
            dict['base_name'] = item[0].strip()
            dict['base_model'] = item[1]
        return super(MetaDataModel, cls).__new__(cls, name, parents, dict)


class DataModelList(six.with_metaclass(MetaDataModel, OrderedDict)):
    ''' Base Class for a list of DataModels '''

    def __init__(self, models=None):

        if models is not None:
            assert all([isinstance(model, self.base_model) for model in models]), \
                'values must be {0} instances.'.format(self.base_name)
            OrderedDict.__init__(self, ((model.release, model) for model in models))
        else:
            OrderedDict.__init__(self, {})

    def __setitem__(self, key, value):
        """Sets a new datamodel."""

        assert isinstance(value, self.base_model), 'value must be a {0}'.format(self.base_name)

        super(DataModelList, self).__setitem__(key, value)

    def __getitem__(self, release):
        """Returns model based on release and aliases."""

        if release in self.keys():
            return super(DataModelList, self).__getitem__(release)

        for model in self.values():
            if release in model.aliases:
                return model

        raise KeyError('cannot find release or alias {0!r}'.format(release))

    def __contains__(self, value):
        ''' Returns True based on release/aliases using getitem '''
        try:
            self[value]
            return True
        except KeyError:
            return False

    def __repr__(self):

        return repr([xx for xx in self.values()])

    def add_datamodel(self, dm):
        """Adds a new datamodel. Uses its release as key."""

        assert isinstance(dm, self.base_model), 'value must be a {0}'.format(self.base_name)

        self[dm.release] = dm

    def copy(self):
        return self.__deepcopy__(self)

    def __deepcopy__(self, memo):
        return memo.__class__(models=[copy.deepcopy(model) for model in memo.values()])


class DataModelLookup(object):
    ''' Class for lookups into the Marvin DataModel '''

    def __init__(self, release=None):
        from marvin.utils.datamodel.dap import datamodel as dap_dms
        from marvin.utils.datamodel.drp import datamodel as drp_cube_dms
        from marvin.utils.datamodel.drp import datamodel_rss as drp_rss_dms
        from marvin.utils.datamodel.query import datamodel as query_dms
        from marvin.utils.datamodel.vacs import datamodel as vac_dms

        self.release = release if release else config.release
        assert release in query_dms.keys(), 'release must be a valid MPL'

        # set datamodels for a given release
        self.dapdm = dap_dms[release]
        self.drpcubedm = drp_cube_dms[release]
        self.drprssdm = drp_rss_dms[release]
        self.querydm = query_dms[release]
        self.vacdm = vac_dms[release] if release in vac_dms else None
        self._dm = None
        self.model_map = ['drp_cube', 'drp_rss', 'dap', 'query', 'vac']

    def __repr__(self):

        return ('<DataModelLookup release={0!r}>'.format(self.release))

    def check_value(self, value, model=None):
        ''' Check that a value is in the Marvin datamodel

        Searches a specified datamodel for a value. If no model is specified,
        attempts to search all the datamodels.

        Parameters:
            value (str):
                The value to look up in the datamodels
            model (str):
                Optional name of the datamodel to search on.
                Can be drp_cube, drp_rss, dap, or query.

        Returns:
            True if value found within a specified model.  When no model is specified,
            returns the name of the model the value was found in.

        '''

        assert isinstance(value, six.string_types), 'value must be a string'
        assert model in self.model_map + [None], 'model must be drp_cube, drp_rss, dap, or query'

        indrpcube = value in self.drpcubedm
        indrprss = value in self.drprssdm
        indap = value in self.dapdm
        inquery = value in self.querydm
        invac = value in self.vacdm if self.vacdm else False
        true_map = [indrpcube, indrprss, indap, inquery, invac]

        if model == 'dap':
            self._dm = self.dapdm
            return indap
        elif model == 'drp_cube':
            self._dm = self.drpcubedm
            return indrpcube
        elif model == 'drp_rss':
            self._dm = self.drprssdm
            return indrprss
        elif model == 'query':
            self._dm = self.querydm
            return inquery
        elif model == 'vac':
            self._dm = self.vacdm
        else:
            # check all of them
            tootrue = sum([indrpcube, indrprss, indap]) > 1
            if tootrue:
                subset = [i for i in self.model_map if true_map[self.model_map.index(i)]]
                raise ValueError('{0} found in multiple datamodels {1}. '
                                 'Fine-tune your value or try a specific model'.format(value,
                                                                                       subset))
            else:
                model = 'drp_cube' if indrpcube else 'drp_rss' if indrprss else \
                    'dap' if indap else 'query' if inquery else 'vac' if invac else None

            if not model:
                print('{0} not found in any datamodels. '
                      'Refine your syntax or check the MaNGA TRM!'.format(value))

            return model

    def get_value(self, value, model=None):
        ''' Get the property or parameter for a given value

        Parameters:
            value (str):
                The name of the value to retrieve
            model (str):
                The name of the model to get the value from

        Returns:
            The parameter or property from the datamodel

        '''

        # check the value
        checked = self.check_value(value, model=model)

        if checked is True:
            param = value == self._dm
        elif checked == 'drp_cube':
            param = value == self.drpcubedm
        elif checked == 'drp_rss':
            param = value == self.drprssdm
        elif checked == 'dap':
            param = value == self.dapdm
        elif checked == 'query':
            param = value == self.querydm
        elif checked == 'vac':
            param = value == self.vacdm
        elif not checked:
            print('No matching parameter found for {0} in model {1}'.format(value, model))
            param = None

        return param

    def write_csv(self, path=None, filename=None, model=None, overwrite=None, **kwargs):
        ''' Writes the datamodels out to CSV '''

        assert model in self.model_map + [None], 'model must be drp, dap, query, or vac'

        if model == 'query':
            self.querydm.write_csv(path=path, filename=filename,
                                   overwrite=overwrite, db=True)
        elif model == 'dap':
            self.dapdm.properties.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
            self.dapdm.models.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
        elif model == 'drp_cube':
            self.drpcubedm.spectra.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
            self.drpcubedm.datacubes.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
        elif model == 'drp_rss':
            self.drprssdm.spectra.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
            self.drprssdm.rss.write_csv(path=path, filename=filename, overwrite=overwrite, **kwargs)
        elif model == 'vac':
            pass

    def write_csvs(self, overwrite=None, **kwargs):
        ''' Write out all models to CSV files '''

        for model in self.model_map:
            self.write_csv(model=model, overwrite=overwrite, **kwargs)
